# Standard library imports.
import sys
import os

# Major package imports.
import wx
import numpy as np

# Enthought library imports.
from traits.api import Any, Bool, Int, Str, List, Event, Enum, Instance, File, Unicode, Property, provides, on_trait_change
from pyface.key_pressed_event import KeyPressedEvent

# Local imports.
from omnivore.framework.editor import FrameworkEditor
from omnivore.framework.document import Document
from omnivore.framework.actions import *
from grid_control import HexEditControl
from omnivore.utils.file_guess import FileMetadata
from omnivore.arch.machine import Machine, Atari800
from omnivore.utils.segmentutil import DefaultSegment, AnticFontSegment
from omnivore.utils.searchutil import known_searchers
from omnivore.utils.processutil import run_detach

from actions import *
from commands import PasteCommand

import logging
log = logging.getLogger(__name__)


class HexEditor(FrameworkEditor):
    """ The toolkit specific implementation of a HexEditor.  See the
    IHexEditor interface for the API documentation.
    """

    #### 'IPythonEditor' interface ############################################

    obj = Instance(File)

    #### traits
    
    grid_range_selected = Bool
    
    segment_parser = Any
    
    segment_number = Int(0)
    
    emulator_label = Unicode("Run Emulator")

    segment_parser_label = Unicode("<parser type>")
    
    ### View traits
    
    map_width = Int
    
    map_zoom = Int
    
    bitmap_width = Int
    
    bitmap_zoom = Int
    
    machine = Any
    
    segment = Any(None)
    
    last_cursor_index = Int(0)
    
    last_anchor_start_index = Int(0)
    
    last_anchor_end_index = Int(0)
    
    can_copy_baseline = Bool

    #### Events ####

    changed = Event

    key_pressed = Event(KeyPressedEvent)
    
    # Class attributes (not traits)
    
    searchers = known_searchers
    
    rect_select = False
    
    ##### Default traits
    
    def _machine_default(self):
        return Atari800
    
    def _segment_default(self):
        return DefaultSegment()
    
    def _map_width_default(self):
        prefs = self.task.get_preferences()
        return prefs.map_width
    
    def _map_zoom_default(self):
        return 2
    
    def _bitmap_width_default(self):
        prefs = self.task.get_preferences()
        return prefs.bitmap_width
    
    def _bitmap_zoom_default(self):
        return 5

    ###########################################################################
    # 'FrameworkEditor' interface.
    ###########################################################################

    def create(self, parent):
        self.control = self._create_control(parent)
        self.machine.one_time_init(self)
        self.task.machine_menu_changed = self.machine

    def load_builtin_extra_metadata(self, doc):
        """ see if the document matches any hardcoded document information
        """
        from omnivore.utils.extra_metadata import check_builtin
        e = check_builtin(doc)
        if 'machine mime' not in e:
            e['machine mime'] = doc.metadata.mime
        return e

    def get_filesystem_extra_metadata_uri(self, doc):
        return doc.metadata.uri + ".omnivore"

    def process_extra_metadata(self, doc, e):
        """ Set up any pre-calculated segments based on the type or content of
        the just-loaded document.
        """
        if 'machine mime' in e:
            mime = e['machine mime']
            if not mime.startswith(self.machine.mime_prefix):
                m = self.machine.find_machine_by_mime(mime)
                if m is not None:
                    self.machine = m
        if 'user segments' in e:
            for s in e['user segments']:
                doc.add_user_segment(s, replace=True)
        first_segment = doc.segments[0]
        if 'serialized user segments' in e:
            for s in e['serialized user segments']:
                s.reconstruct_raw(first_segment.rawdata) 
                doc.add_user_segment(s, replace=True)
        first_segment.restore_extra_from_dict(e)
        if 'emulator' in e:
            doc.emulator = e['emulator']
        if 'font' in e:
            # FIXME: I don't think 'font' is set anywhere, so this never gets called
            self.machine.set_font(e['font'][0], e['font'][1])
        if 'initial segment' in e:
            self.initial_segment = e['initial segment']
        if 'baseline document' in e:
            self.load_baseline(e['baseline document'], doc, ignore_error=True)
        if not doc.has_baseline:
            self.use_self_as_baseline(doc)
        if 'diff highlight' in e:
            self.diff_highlight = doc.has_baseline and bool(e['diff highlight'])
        if 'map width' in e:
            self.map_width = e['map width']
        if 'map zoom' in e:
            self.map_zoom = e['map zoom']
        if 'bitmap width' in e:
            self.bitmap_width = e['bitmap width']
        if 'bitmap zoom' in e:
            self.bitmap_zoom = e['bitmap zoom']
        self.machine.restore_extra_from_dict(e)
    
    def get_extra_metadata(self, mdict):
        mdict["serialized user segments"] = list(self.document.user_segments)
        base = self.document.segments[0]
        base.serialize_extra_to_dict(mdict)

        emu = self.document.emulator
        if emu and not 'system default' in emu:
            mdict["emulator"] = self.document.emulator
        if self.document.baseline_document is not None:
            mdict["baseline document"] = self.document.baseline_document.metadata.uri
        mdict["diff highlight"] = self.diff_highlight
        mdict["map width"] = self.map_width
        mdict["map zoom"] = self.map_zoom
        mdict["bitmap width"] = self.bitmap_width
        mdict["bitmap zoom"] = self.bitmap_zoom
        self.machine.serialize_extra_to_dict(mdict)

    def rebuild_document_properties(self):
        FrameworkEditor.rebuild_document_properties(self)
        self.find_segment()
        self.update_emulator()
        self.compare_to_baseline()
        self.task.machine_menu_changed = self.machine
    
    def copy_view_properties(self, old_editor):
        try:
            self.machine = old_editor.machine.clone_machine()
        except AttributeError:
            self.machine = self._machine_default()
    
    @property
    def document_length(self):
        return len(self.segment)
    
    def process_paste_data_object(self, data_obj, cmd_cls=None):
        bytes, extra = self.get_numpy_from_data_object(data_obj)
        ranges, indexes = self.get_selected_ranges_and_indexes()
        if extra and extra[0] == "numpy,multiple":
            source_indexes = extra[1]
        else:
            source_indexes = None
        if cmd_cls is None:
            cmd_cls = PasteCommand
        cmd = cmd_cls(self.segment, ranges, self.cursor_index, bytes, source_indexes)
        self.process_command(cmd)
    
    def get_numpy_from_data_object(self, data_obj):
        # Full list of valid data formats:
        #
        # >>> import wx
        # >>> [x for x in dir(wx) if x.startswith("DF_")]
        # ['DF_BITMAP', 'DF_DIB', 'DF_DIF', 'DF_ENHMETAFILE', 'DF_FILENAME',
        # 'DF_HTML', 'DF_INVALID', 'DF_LOCALE', 'DF_MAX', 'DF_METAFILE',
        # 'DF_OEMTEXT', 'DF_PALETTE', 'DF_PENDATA', 'DF_PRIVATE', 'DF_RIFF',
        # 'DF_SYLK', 'DF_TEXT', 'DF_TIFF', 'DF_UNICODETEXT', 'DF_WAVE']
        extra = None
        if wx.DF_TEXT in data_obj.GetAllFormats():
            value = data_obj.GetText().encode('utf-8')
        elif wx.DF_UNICODETEXT in data_obj.GetAllFormats():  # for windows
            value = data_obj.GetText().encode('utf-8')
        else:
            value = data_obj.GetData()
            fmt = data_obj.GetPreferredFormat()
            if fmt.GetId() == "numpy,columns":
                r, c, value = value.split(",", 2)
                extra = fmt.GetId(), int(r), int(c)
            elif fmt.GetId() == "numpy,multiple":
                i, value = value.split(",", 1)
                i = int(i)
                value, index_string = value[0:i], value[i:]
                indexes = np.fromstring(index_string, dtype=np.uint32)
                extra = fmt.GetId(), indexes
        bytes = np.fromstring(value, dtype=np.uint8)
        return bytes, extra
    
    supported_clipboard_data_objects = [
        wx.CustomDataObject("numpy,multiple"),
        wx.CustomDataObject("numpy"),
        wx.CustomDataObject("numpy,columns"),
        wx.TextDataObject(),
        ]
    
    def create_clipboard_data_object(self):
        ranges, indexes = self.get_selected_ranges_and_indexes()
        if len(ranges) == 1:
            r = ranges[0]
            data = self.segment[r[0]:r[1]]
            data_obj = wx.CustomDataObject("numpy")
            data_obj.SetData(data.tostring())
        elif np.alen(indexes) > 0:
            data = self.segment[indexes]
            s1 = data.tostring()
            s2 = indexes.tostring()
            data_obj = wx.CustomDataObject("numpy,multiple")
            s = "%d,%s%s" % (len(s1), s1, s2)
            data_obj.SetData(s)
        else:
            data_obj = None
        if data_obj is not None:
            text = " ".join(["%02x" % i for i in data])
            text_obj = wx.TextDataObject()
            text_obj.SetText(text)
            c = wx.DataObjectComposite()
            c.Add(data_obj)
            c.Add(text_obj)
            return c
        return None

    def update_panes(self):
        self.segment = self.document.segments[self.segment_number]
        self.reconfigure_panes()
        self.update_segments_ui()
    
    def update_segments_ui(self):
        self.segment_list.set_segments(self.document.segments, self.segment_number)
        if self.segment_parser is not None:
            self.segment_parser_label = self.segment_parser.menu_name
        else:
            self.segment_parser_label = "No parser"
        self.task.segments_changed = self.document.segments
        self.task.segment_selected = self.segment_number
    
    def reconfigure_panes(self):
        self.hex_edit.recalc_view()
        self.disassembly.recalc_view()
        self.bitmap.recalc_view()
        self.font_map.recalc_view()
    
    def check_document_change(self):
        if self.last_cursor_index != self.cursor_index or self.last_anchor_start_index != self.anchor_start_index or self.last_anchor_end_index != self.anchor_end_index:
            self.document.change_count += 1
            self.update_cursor_history()

    def get_cursor_state(self):
        return self.segment, self.cursor_index

    def restore_cursor_state(self, state):
        segment, index = state
        number = self.document.find_segment_index(segment)
        if number < 0:
            log.error("tried to restore cursor to a deleted segment? %s" % segment)
        else:
            if number != self.segment_number:
                self.view_segment_number(number)
            self.index_clicked(index, 0, None)
        print self.cursor_history
    
    def refresh_panes(self):
        self.check_document_change()
        self.hex_edit.refresh_view()
        self.disassembly.refresh_view()
        self.bitmap.refresh_view()
        self.font_map.refresh_view()
        self.sidebar.refresh_active()
    
    def set_bitmap_width(self, width=None):
        if width is None:
            width = self.bitmap_width
        self.bitmap_width = width
        self.bitmap.recalc_view()
    
    def set_bitmap_zoom(self, zoom=None):
        if zoom is None:
            zoom = self.bitmap_zoom
        self.bitmap_zoom = zoom
        self.bitmap.recalc_view()
    
    @on_trait_change('machine.bitmap_shape_change_event,machine.bitmap_color_change_event')
    def update_bitmap(self):
        self.bitmap.recalc_view()
    
    @on_trait_change('machine.font_change_event')
    def update_fonts(self):
        self.font_map.set_font()
        self.font_map.Refresh()
        pane = self.window.get_dock_pane('hex_edit.font_map')
        pane.name = self.machine.font_mapping.name
        self.window._aui_manager.Update()
    
    @on_trait_change('machine.disassembler_change_event')
    def update_disassembler(self):
        self.disassembly.recalc_view()
    
    @on_trait_change('document.emulator_change_event')
    def update_emulator(self):
        emu = self.document.emulator
        if emu is None:
            emu = self.machine.get_system_default_emulator(self.task)
        if not self.machine.is_known_emulator(emu):
            self.machine.add_emulator(self.task, emu)
        self.emulator_label = "Run using '%s'" % emu['name']

    def run_emulator(self):
        emu = self.document.emulator
        if not emu:
            emu = self.machine.get_system_default_emulator(self.task)
        if self.dirty:
            if not self.save():
                return
        exe = emu['exe']
        args = emu['args']
        fspath = self.document.filesystem_path()
        if fspath is not None:
            try:
                run_detach(exe, args, fspath, "%s")
            except RuntimeError, e:
                self.window.error("Failed launching %s %s\n\nError: %s" % (exe, args, str(e)), "%s Emulator Error" % emu['name'])
        else:
            self.window.error("Can't run emulator on:\n\n%s\n\nDocument is not on local filesystem" % self.document.uri, "%s Emulator Error" % emu['name'])

    def set_map_width(self, width=None):
        if width is None:
            width = self.map_width
        self.map_width = width
        self.font_map.recalc_view()
    
    def set_map_zoom(self, zoom=None):
        if zoom is None:
            zoom = self.map_zoom
        self.map_zoom = zoom
        self.font_map.recalc_view()
    
    def get_font_from_selection(self):
        pass
    
    def set_machine(self, machine):
        self.machine = machine
        self.reconfigure_panes()
    
    def find_segment_parser(self, parsers, segment_name=None):
        self.document.parse_segments(parsers)
        self.find_segment(segment_name)

    def find_first_valid_segment_index(self):
        return 0
    
    def find_segment(self, segment_name=None, segment=None):
        if segment_name is not None:
            index = self.document.find_segment_index_by_name(segment_name)
        elif segment is not None:
            index = self.document.find_segment_index(segment)
        else:
            index = self.find_first_valid_segment_index()
        if index < 0:
            index = 0
        self.segment_number = index
        self.segment_parser = self.document.segment_parser
        self.update_segments_ui()
        self.segment = self.document.segments[index]
        self.view_segment_set_width(self.segment)
        self.select_none(refresh=False)
    
    def set_segment_parser(self, parser):
        self.find_segment_parser([parser])
        self.update_panes()
    
    def view_segment_set_width(self, segment):
        pass
    
    def view_segment_number(self, number):
        doc = self.document
        num = number if number < len(doc.segments) else len(doc.segments) - 1
        if num != self.segment_number:
            old_segment = self.segment
            self.segment = doc.segments[num]
            self.adjust_selection(old_segment)
            self.segment_number = num
            self.invalidate_search()
            self.view_segment_set_width(self.segment)
            self.reconfigure_panes()
            self.task.status_bar.message = "Switched to segment %s" % str(self.segment)
        self.update_segments_ui()
    
    def get_extra_segment_savers(self, segment):
        savers = []
        savers.append(self.disassembly)
        return savers
    
    def save_segment(self, saver, uri):
        try:
            bytes = saver.encode_data(self.segment)
            self.save_to_uri(bytes, uri, save_metadata=False)
        except Exception, e:
            log.error("%s: %s" % (uri, str(e)))
            #self.window.error("Error trying to save:\n\n%s\n\n%s" % (uri, str(e)), "File Save Error")
            raise
    
    def invalidate_search(self):
        self.task.change_minibuffer_editor(self)
    
    def compare_to_baseline(self):
        if self.diff_highlight:
            self.document.update_baseline()
    
    def adjust_selection(self, old_segment):
        """Adjust the selection of the current segment so that it is limited to the
        bounds of the new segment.
        
        If the current selection is entirely out of bounds of the new segment,
        all the selection indexes will be set to zero.
        """
        # find byte index of view into master array
        g = self.document.global_segment
        s = self.segment
        global_offset = g.get_raw_index(0)
        new_offset = s.get_raw_index(0)
        old_offset = old_segment.get_raw_index(0)
        
        self.cursor_index -= new_offset - old_offset
        self.selected_ranges = s.get_style_ranges(selected=True)
        if self.selected_ranges:
            # Arbitrarily puth the anchor on the last selected range
            last = self.selected_ranges[-1]
            self.anchor_initial_start_index = self.anchor_start_index = last[0]
            self.anchor_initial_end_index = self.anchor_end_index = last[1]
        g.clear_style_bits(selected=True)
        self.document.change_count += 1
        self.highlight_selected_ranges()
    
    def highlight_selected_ranges(self):
        s = self.segment
        s.clear_style_bits(selected=True)
        s.set_style_ranges(self.selected_ranges, selected=True)
        self.document.change_count += 1
        self.can_copy_baseline = self.can_copy and self.baseline_present
    
    def convert_ranges(self, from_style, to_style):
        s = self.segment
        ranges = s.get_style_ranges(**from_style)
        s.clear_style_bits(**from_style)
        s.clear_style_bits(**to_style)
        s.set_style_ranges(ranges, **to_style)
        self.selected_ranges = s.get_style_ranges(selected=True)
        self.document.change_count += 1
    
    def get_label_at_index(self, index):
        return self.hex_edit.table.get_label_at_index(index)
    
    def get_label_of_ranges(self, ranges):
        labels = []
        for start, end in ranges:
            labels.append("%s-%s" % (self.get_label_at_index(start), self.get_label_at_index(end)))
        return ", ".join(labels)
    
    def get_segments_from_selection(self, size=-1):
        s = self.segment
        segments = []
        
        # Get the selected ranges directly from the segment style data, because
        # the individual range entries in self.selected_ranges can be out of
        # order or overlapping
        ranges = s.get_style_ranges(selected=True)
        if len(ranges) == 1:
            seg_start, seg_end = ranges[0]
            segment = DefaultSegment(s.rawdata[seg_start:seg_end], s.start_addr + seg_start)
            segments.append(segment)
        elif len(ranges) > 1:
            # If there are multiple selections, use an indexed segment
            indexes = []
            for start, end in ranges:
                indexes.extend(range(start, end))
            raw = s.rawdata.get_indexed(indexes)
            segment = DefaultSegment(raw, s.start_addr + indexes[0])
            segments.append(segment)
        return segments
    
    def get_selected_status_message(self):
        if not self.selected_ranges:
            return ""
        if len(self.selected_ranges) == 1:
            r = self.selected_ranges
            return "[%d bytes selected %s]" % (r[0][1] - r[0][0], self.get_label_of_ranges(r))
        else:
            return "[%d ranges selected]" % (len(self.selected_ranges))
    
    def show_status_message(self, msg):
        s = self.get_selected_status_message()
        if s:
            msg = "%s %s" % (s, msg)
        self.task.status_bar.message = msg
    
    def add_user_segment(self, segment, update=True):
        self.document.add_user_segment(segment)
        if update:
            self.update_segments_ui()
            self.segment_list.ensure_visible(segment)
        self.metadata_dirty = True
    
    def delete_user_segment(self, segment):
        self.document.delete_user_segment(segment)
        self.view_segment_number(self.segment_number)
        self.metadata_dirty = True
    
    def find_in_user_segment(self, base_index):
        for s in self.document.user_segments:
            try:
                index = s.get_index_from_base_index(base_index)
                return s, index
            except IndexError:
                continue
        return None, None
    
    def ensure_visible(self, start, end):
        self.index_clicked(start, 0, None)
    
    def update_history(self):
#        history = document.undo_stack.serialize()
#        self.window.application.save_log(str(history), "command_log", ".log")
        self.undo_history.update_history()

    def mark_index_range_changed(self, index_range):
        self.disassembly.restart_disassembly(index_range[0])
    
    def perform_idle(self):
        self.disassembly.perform_idle()
    
    def common_popup_actions(self):
        return [CutAction, CopyAction, CopyDisassemblyAction, CopyAsReprAction, PasteAction, None, SelectAllAction, SelectNoneAction, GetSegmentFromSelectionAction, None, MarkSelectionAsCodeAction, MarkSelectionAsDataAction, MarkSelectionAsDisplayListAction, MarkSelectionAsJumpmanLevelAction, MarkSelectionAsJumpmanHarvestAction, RevertToBaselineAction, None, AddCommentPopupAction, RemoveCommentPopupAction]
    
    def change_bytes(self, start, end, bytes, pretty=None):
        """Convenience function to perform a ChangeBytesCommand
        """
        print "changing bytes %d-%d to %s" % (start, end, repr(bytes))
        self.document.change_count += 1
        cmd = CoalescingChangeByteCommand(self.segment, start, end, bytes)
        if pretty:
            cmd.pretty_name = pretty
        self.process_command(cmd)

    ###########################################################################
    # Trait handlers.
    ###########################################################################


    ###########################################################################
    # Private interface.
    ###########################################################################

    def _create_control(self, parent):
        """ Creates the toolkit-specific control for the widget. """

        # Base-class constructor.
        self.hex_edit = HexEditControl(parent, self.task)

        ##########################################
        # Events.
        ##########################################

        # Get related controls
        self.disassembly = self.window.get_dock_pane('hex_edit.disassembly').control
        self.bitmap = self.window.get_dock_pane('hex_edit.bitmap').control
        self.font_map = self.window.get_dock_pane('hex_edit.font_map').control
        self.segment_list = self.window.get_dock_pane('hex_edit.segments').control
        self.undo_history = self.window.get_dock_pane('hex_edit.undo').control
        self.sidebar = self.window.get_dock_pane('hex_edit.sidebar')

        # Load the editor's contents.
        self.load()

        return self.hex_edit

    #### wx event handlers ####################################################
    
    def index_clicked(self, index, bit, control):
        self.cursor_index = index
        self.check_document_change()
        if control != self.hex_edit:
            self.hex_edit.select_index(index)
        if control != self.disassembly:
            self.disassembly.select_index(index)
        if control != self.bitmap:
            self.bitmap.select_index(index)
        if control != self.font_map:
            self.font_map.select_index(index)
        self.sidebar.refresh_active()
        self.can_copy = (self.anchor_start_index != self.anchor_end_index)
        self.can_copy_baseline = self.can_copy and self.baseline_present



##########################################################################
#
# Processor specific code

# CPU = "65816"
# Description = "Western Design Center 65816 8/16-bit microprocessor."
# DataWidth = 8/16  # 8-bit or 16-bit data depending on processor mode
# AddressWidth = 16/24  # 16-bit or 24-bit addresses depending on processor mode

# Maximum length of an instruction (for formatting purposes)
maxLength = 4

# Leadin bytes for multibyte instructions
leadInBytes = []

# Addressing mode table
# List of addressing modes and corresponding format strings for operands.
addressModeTable = {
"implicit"                : "",
"absolute"                : "${1:02X}{0:02X}",
"absolutex"               : "${1:02X}{0:02X},x",
"absolutey"               : "${1:02X}{0:02X},y",
"accumulator"             : "a",
"immediate"               : "#${0:02X}",
"indirectx"               : "(${0:02X},x)",
"indirecty"               : "(${0:02X}),y",
"indirect"                : "(${1:02X}{0:02X})",
"relative"                : "${0:04X}",
"relativelong"            : "${0:04X}",
"zeropage"                : "${0:02X}",
"zeropagex"               : "${0:02X},x",
"zeropagey"               : "${0:02X},y",
"indirectzeropage"        : "(${0:02X})",
"absoluteindexedindirect" : "(${1:02X}{0:02X},x)",
"stackrelative"           : "${0:02X},s",
"stackrelativeindirecty"  : "(${0:02X},s),y",
"absolutelong"            : "${2:02X}{1:02X}{0:02X}",
"absolutelongx"           : "${2:02X}{1:02X}{0:02X},x",
"absoluteindirectx"       : "(${1:02X}{0:02X},x)",
"absoluteindirectlong"    : "[${1:02X}{0:02X}]",
"directpageindirect"      : "(${0:02X})",
"directpageindirectlong"  : "[${0:02X}]",
"directpageindirectlongy" : "[${0:02X}],y",
"blockmove"               : "${0:02X},${1:02X}",
}

# Op Code Table
# Key is numeric opcode (possibly multiple bytes)
# Value is a list:
#   # bytes
#   mnemonic
#   addressing mode
#   flags (e.g. pcr)

# regex conversion from http://www.zophar.net/fileuploads/2/10538ivwiu/65816info.txt
# find:    ([0-9A-F]+)      ([A-Z]+)     (.+?)( +)([1-4])
# replace: 0x\1 : [ \5, "\2",  "\3"\4],
opcodeTable = {
0x00 : [ 1, "brk",  "implicit"                ],
0x01 : [ 2, "ora",  "indirectx"               ],
0x02 : [ 2, "cop",  "zeropage"                ],
0x03 : [ 2, "ora",  "stackrelative"           ],
0x04 : [ 2, "tsb",  "zeropage"                ],
0x05 : [ 2, "ora",  "zeropage"                ],
0x06 : [ 2, "asl",  "zeropage"                ],
0x07 : [ 2, "ora",  "directpageindirectlong"  ],
0x08 : [ 1, "php",  "implicit"                ],
0x09 : [ 2, "ora",  "immediate"               ],
0x0a : [ 1, "asl",  "accumulator"             ],
0x0b : [ 1, "phd",  "implicit"                ],
0x0c : [ 3, "tsb",  "absolute"                ],
0x0d : [ 3, "ora",  "absolute"                ],
0x0e : [ 3, "asl",  "absolute"                ],
0x0f : [ 4, "ora",  "absolutelong"            ],

0x10 : [ 2, "bpl",  "relative", pcr           ],
0x11 : [ 2, "ora",  "indirecty"               ],
0x12 : [ 2, "ora",  "indirectzeropage"        ],
0x13 : [ 2, "ora",  "stackrelativeindirecty"  ],
0x14 : [ 2, "trb",  "zeropage"                ],
0x15 : [ 2, "ora",  "zeropagex"               ],
0x16 : [ 2, "asl",  "zeropagex"               ],
0x17 : [ 2, "ora",  "directpageindirectlongy" ],
0x18 : [ 1, "clc",  "implicit"                ],
0x19 : [ 3, "ora",  "absolutey"               ],
0x1a : [ 1, "inc",  "accumulator"             ],
0x1b : [ 1, "tcs",  "implicit"                ],
0x1c : [ 3, "trb",  "absolute"                ],
0x1d : [ 3, "ora",  "absolutex"               ],
0x1e : [ 3, "asl",  "absolutex"               ],
0x1f : [ 4, "ora",  "absolutelongx"           ],

0x20 : [ 3, "jsr",  "absolute"                ],
0x21 : [ 2, "and",  "indirectx"               ],
0x22 : [ 4, "jsr",  "absolutelong"            ],
0x23 : [ 2, "and",  "stackrelative"           ],
0x24 : [ 2, "bit",  "zeropage"                ],
0x25 : [ 2, "and",  "zeropage"                ],
0x26 : [ 2, "rol",  "zeropage"                ],
0x27 : [ 2, "and",  "directpageindirectlong"  ],
0x28 : [ 1, "plp",  "implicit"                ],
0x29 : [ 2, "and",  "immediate"               ],
0x2a : [ 1, "rol",  "accumulator"             ],
0x2b : [ 1, "pld",  "implicit"                ],
0x2c : [ 3, "bit",  "absolute"                ],
0x2d : [ 3, "and",  "absolute"                ],
0x2e : [ 3, "rol",  "absolute"                ],
0x2f : [ 4, "and",  "absolutelong"            ],

0x30 : [ 2, "bmi",  "relative", pcr           ],
0x31 : [ 2, "and",  "indirecty"               ],
0x32 : [ 2, "and",  "indirectzeropage"        ],
0x33 : [ 2, "and",  "stackrelativeindirecty"  ],
0x34 : [ 2, "bit",  "zeropagex"               ],
0x35 : [ 2, "and",  "zeropagex"               ],
0x36 : [ 2, "rol",  "zeropagex"               ],
0x37 : [ 2, "and",  "directpageindirectlongy" ],
0x38 : [ 1, "sec",  "implicit"                ],
0x39 : [ 3, "and",  "absolutey"               ],
0x3a : [ 1, "dec",  "accumulator"             ],
0x3b : [ 1, "tsc",  "implicit"                ],
0x3c : [ 3, "bit",  "absolutex"               ],
0x3d : [ 3, "and",  "absolutex"               ],
0x3e : [ 3, "rol",  "absolutex"               ],
0x3f : [ 4, "and",  "absolutelongx"           ],

0x40 : [ 1, "rti",  "implicit"                ],
0x41 : [ 2, "eor",  "indirectx"               ],
0x42 : [ 2, "wdm",  "zeropage"                ],
0x43 : [ 2, "eor",  "stackrelative"           ],
0x44 : [ 3, "mvp",  "blockmove"               ],
0x45 : [ 2, "eor",  "zeropage"                ],
0x46 : [ 2, "lsr",  "zeropage"                ],
0x47 : [ 2, "eor",  "directpageindirectlong"  ],
0x48 : [ 1, "pha",  "implicit"                ],
0x49 : [ 2, "eor",  "immediate"               ],
0x4a : [ 1, "lsr",  "accumulator"             ],
0x4b : [ 1, "phk",  "implicit"                ],
0x4c : [ 3, "jmp",  "absolute"                ],
0x4d : [ 3, "eor",  "absolute"                ],
0x4e : [ 3, "lsr",  "absolute"                ],
0x4f : [ 4, "eor",  "absolutelong"            ],

0x50 : [ 2, "bvc",  "relative", pcr           ],
0x51 : [ 2, "eor",  "indirecty"               ],
0x52 : [ 2, "eor",  "indirectzeropage"        ],
0x53 : [ 2, "eor",  "stackrelativeindirecty"  ],
0x54 : [ 3, "mvn",  "blockmove"               ],
0x55 : [ 2, "eor",  "zeropagex"               ],
0x56 : [ 2, "lsr",  "zeropagex"               ],
0x57 : [ 2, "eor",  "directpageindirectlongy" ],
0x58 : [ 1, "cli",  "implicit"                ],
0x59 : [ 3, "eor",  "absolutey"               ],
0x5a : [ 1, "phy",  "implicit"                ],
0x5b : [ 1, "tcd",  "implicit"                ],
0x5c : [ 4, "jmp",  "absolutelong"            ],
0x5d : [ 3, "eor",  "absolutex"               ],
0x5e : [ 3, "lsr",  "absolutex"               ],
0x5f : [ 4, "eor",  "absolutelongx"           ],

0x60 : [ 1, "rts",  "implicit"                ],
0x61 : [ 2, "adc",  "indirectx"               ],
0x62 : [ 3, "per",  "absolute"                ],
0x63 : [ 2, "adc",  "stackrelative"           ],
0x64 : [ 2, "stz",  "zeropage"                ],
0x65 : [ 2, "adc",  "zeropage"                ],
0x66 : [ 2, "ror",  "zeropage"                ],
0x67 : [ 2, "adc",  "directpageindirectlong"  ],
0x68 : [ 1, "pla",  "implicit"                ],
0x69 : [ 2, "adc",  "immediate"               ],
0x6a : [ 1, "ror",  "accumulator"             ],
0x6b : [ 1, "rtl",  "implicit"                ],
0x6c : [ 3, "jmp",  "indirect"                ],
0x6d : [ 3, "adc",  "absolute"                ],
0x6e : [ 3, "ror",  "absolute"                ],
0x6f : [ 4, "adc",  "absolutelong"            ],

0x70 : [ 2, "bvs",  "relative", pcr           ],
0x71 : [ 2, "adc",  "indirecty"               ],
0x72 : [ 2, "adc",  "indirectzeropage"        ],
0x73 : [ 2, "adc",  "stackrelativeindirecty"  ],
0x74 : [ 2, "stz",  "zeropagex"               ],
0x74 : [ 2, "stz",  "zeropagex"               ],
0x75 : [ 2, "adc",  "zeropagex"               ],
0x76 : [ 2, "ror",  "zeropagex"               ],
0x77 : [ 2, "adc",  "directpageindirectlongy" ],
0x78 : [ 1, "sei",  "implicit"                ],
0x79 : [ 3, "adc",  "absolutey"               ],
0x7a : [ 1, "ply",  "implicit"                ],
0x7b : [ 1, "tdc",  "implicit"                ],
0x7c : [ 3, "jmp",  "absoluteindexedindirect" ],
0x7d : [ 3, "adc",  "absolutex"               ],
0x7e : [ 3, "ror",  "absolutex"               ],
0x7f : [ 4, "adc",  "absolutelongx"           ],

0x80 : [ 2, "bra",  "relative", pcr           ],
0x81 : [ 2, "sta",  "indirectx"               ],
0x82 : [ 3, "brl",  "relativelong", pcr       ],
0x83 : [ 2, "sta",  "stackrelative"           ],
0x84 : [ 2, "sty",  "zeropage"                ],
0x85 : [ 2, "sta",  "zeropage"                ],
0x86 : [ 2, "stx",  "zeropage"                ],
0x87 : [ 2, "sta",  "directpageindirectlong"  ],
0x88 : [ 1, "dey",  "implicit"                ],
0x89 : [ 2, "bit",  "immediate"               ],
0x8a : [ 1, "txa",  "implicit"                ],
0x8b : [ 1, "phb",  "implicit"                ],
0x8c : [ 3, "sty",  "absolute"                ],
0x8d : [ 3, "sta",  "absolute"                ],
0x8e : [ 3, "stx",  "absolute"                ],
0x8f : [ 4, "sta",  "absolutelong"            ],

0x90 : [ 2, "bcc",  "relative", pcr           ],
0x91 : [ 2, "sta",  "indirecty"               ],
0x92 : [ 2, "sta",  "indirectzeropage"        ],
0x93 : [ 2, "sta",  "stackrelativeindirecty"  ],
0x94 : [ 2, "sty",  "zeropagex"               ],
0x95 : [ 2, "sta",  "zeropagex"               ],
0x96 : [ 2, "stx",  "zeropagey"               ],
0x97 : [ 2, "sta",  "directpageindirectlongy" ],
0x98 : [ 1, "tya",  "implicit"                ],
0x99 : [ 3, "sta",  "absolutey"               ],
0x9a : [ 1, "txs",  "implicit"                ],
0x9b : [ 1, "txy",  "implicit"                ],
0x9c : [ 3, "stz",  "absolute"                ],
0x9d : [ 3, "sta",  "absolutex"               ],
0x9e : [ 3, "stz",  "absolutex"               ],
0x9f : [ 4, "sta",  "absolutelongx"           ],

0xa0 : [ 2, "ldy",  "immediate"               ],
0xa1 : [ 2, "lda",  "indirectx"               ],
0xa2 : [ 2, "ldx",  "immediate"               ],
0xa3 : [ 2, "lda",  "stackrelative"           ],
0xa4 : [ 2, "ldy",  "zeropage"                ],
0xa5 : [ 2, "lda",  "zeropage"                ],
0xa6 : [ 2, "ldx",  "zeropage"                ],
0xa7 : [ 2, "lda",  "directpageindirectlong"  ],
0xa8 : [ 1, "tay",  "implicit"                ],
0xa9 : [ 2, "lda",  "immediate"               ],
0xaa : [ 1, "tax",  "implicit"                ],
0xab : [ 1, "plb",  "implicit"                ],
0xac : [ 3, "ldy",  "absolute"                ],
0xad : [ 3, "lda",  "absolute"                ],
0xae : [ 3, "ldx",  "absolute"                ],
0xaf : [ 4, "lda",  "absolutelong"            ],

0xb0 : [ 2, "bcs",  "relative", pcr           ],
0xb1 : [ 2, "lda",  "indirecty"               ],
0xb2 : [ 2, "lda",  "indirectzeropage"        ],
0xb3 : [ 2, "lda",  "stackrelativeindirecty"  ],
0xb4 : [ 2, "ldy",  "zeropagex"               ],
0xb5 : [ 2, "lda",  "zeropagex"               ],
0xb6 : [ 2, "ldx",  "zeropagey"               ],
0xb7 : [ 2, "lda",  "directpageindirectlongy" ],
0xb8 : [ 1, "clv",  "implicit"                ],
0xb9 : [ 3, "lda",  "absolutey"               ],
0xba : [ 1, "tsx",  "implicit"                ],
0xbb : [ 1, "tyx",  "implicit"                ],
0xbc : [ 3, "ldy",  "absolutex"               ],
0xbd : [ 3, "lda",  "absolutex"               ],
0xbe : [ 3, "ldx",  "absolutey"               ],
0xbf : [ 4, "lda",  "absolutelongx"           ],

0xc0 : [ 2, "cpy",  "immediate"               ],
0xc1 : [ 2, "cmp",  "indirectx"               ],
0xc2 : [ 2, "rep",  "immediate"               ],
0xc3 : [ 2, "cmp",  "stackrelative"           ],
0xc4 : [ 2, "cpy",  "zeropage"                ],
0xc5 : [ 2, "cmp",  "zeropage"                ],
0xc6 : [ 2, "dec",  "zeropage"                ],
0xc7 : [ 2, "cmp",  "directpageindirectlong"  ],
0xc8 : [ 1, "iny",  "implicit"                ],
0xc9 : [ 2, "cmp",  "immediate"               ],
0xca : [ 1, "dex",  "implicit"                ],
0xcb : [ 1, "wai",  "implicit"                ],
0xcc : [ 3, "cpy",  "absolute"                ],
0xcd : [ 3, "cmp",  "absolute"                ],
0xce : [ 3, "dec",  "absolute"                ],
0xcf : [ 4, "cmp",  "absolutelong"            ],

0xd0 : [ 2, "bne",  "relative", pcr           ],
0xd1 : [ 2, "cmp",  "indirecty"               ],
0xd2 : [ 2, "cmp",  "indirectzeropage"        ],
0xd3 : [ 2, "cmp",  "stackrelativeindirecty"  ],
0xd4 : [ 2, "pei",  "directpageindirect"      ],
0xd5 : [ 2, "cmp",  "zeropagex"               ],
0xd6 : [ 2, "dec",  "zeropagex"               ],
0xd7 : [ 2, "cmp",  "directpageindirectlongy" ],
0xd8 : [ 1, "cld",  "implicit"                ],
0xd9 : [ 3, "cmp",  "absolutey"               ],
0xda : [ 1, "phx",  "implicit"                ],
0xdb : [ 1, "stp",  "implicit"                ],
0xdc : [ 3, "jmp",  "absoluteindirectlong"    ],
0xdd : [ 3, "cmp",  "absolutex"               ],
0xde : [ 3, "dec",  "absolutex"               ],
0xdf : [ 4, "cmp",  "absolutelongx"           ],

0xe0 : [ 2, "cpx",  "immediate"               ],
0xe1 : [ 2, "sbc",  "indirectx"               ],
0xe2 : [ 2, "sep",  "immediate"               ],
0xe3 : [ 2, "sbc",  "stackrelative"           ],
0xe4 : [ 2, "cpx",  "zeropage"                ],
0xe5 : [ 2, "sbc",  "zeropage"                ],
0xe6 : [ 2, "inc",  "zeropage"                ],
0xe7 : [ 2, "sbc",  "directpageindirectlong"  ],
0xe8 : [ 1, "inx",  "implicit"                ],
0xe9 : [ 2, "sbc",  "immediate"               ],
0xea : [ 1, "nop",  "implicit"                ],
0xeb : [ 1, "xba",  "implicit"                ],
0xec : [ 3, "cpx",  "absolute"                ],
0xed : [ 3, "sbc",  "absolute"                ],
0xee : [ 3, "inc",  "absolute"                ],
0xef : [ 4, "sbc",  "absolutelong"            ],

0xf0 : [ 2, "beq",  "relative", pcr           ],
0xf1 : [ 2, "sbc",  "indirecty"               ],
0xf2 : [ 2, "sbc",  "indirectzeropage"        ],
0xf3 : [ 2, "sbc",  "stackrelativeindirecty"  ],
0xf4 : [ 3, "pea",  "absolute"                ],
0xf5 : [ 2, "sbc",  "zeropagex"               ],
0xf6 : [ 2, "inc",  "zeropagex"               ],
0xf7 : [ 2, "sbc",  "directpageindirectlongy" ],
0xf8 : [ 1, "sed",  "implicit"                ],
0xf9 : [ 3, "sbc",  "absolutey"               ],
0xfa : [ 1, "plx",  "implicit"                ],
0xfb : [ 1, "xce",  "implicit"                ],
0xfd : [ 3, "sbc",  "absolutex"               ],
0xfe : [ 3, "inc",  "absolutex"               ],
0xff : [ 4, "sbc",  "absolutelongx"           ],
}

# End of processor specific code
##########################################################################



import os
import types
import cStringIO as StringIO

import numpy as np
import fs

# Enthought library imports.
from traits.api import HasTraits, Trait, TraitHandler, Int, Any, List, Set, Bool, Event, Dict, Set, Unicode, Property, Str

from atrcopy import SegmentData, DefaultSegment, DefaultSegmentParser, InvalidSegmentParser

from omnivore.utils.command import UndoStack
from omnivore.utils.file_guess import FileGuess, FileMetadata


class TraitNumpyConverter(TraitHandler):
    """Trait validator to convert bytes to numpy array"""
    def validate(self, object, name, value):
        if type(value) is np.ndarray:
            return value
        elif type(value) is types.StringType:
            return np.fromstring(value, dtype=np.uint8)
        self.error(object, name, value)

    def info(self):
        return '**a string or numpy array**'


class Document(HasTraits):
    undo_stack = Any
    
    metadata = Any

    name = Property(Unicode, depends_on='metadata')

    uri = Property(Unicode, depends_on='metadata')

    read_only = Property(Bool, depends_on='metadata')

    document_id = Int(-1)
    
    baseline_document = Any(transient=True)
    
    last_task_id = Str
    
    bytes = Trait("", TraitNumpyConverter())
    
    style = Trait("", TraitNumpyConverter())
    
    segments = List
    
    segment_parser = Any
    
    user_segments = List
    
    extra_metadata = Dict
    
    emulator = Any

    emulator_change_event = Event
    
    # Trait events to provide view updating
    
    undo_stack_changed = Event
    
    byte_values_changed = Event  # but not the size of the bytes array. That's not handled yet
    
    change_count = Int()
    
    can_revert = Property(Bool, depends_on='metadata')

    permute = Any

    #### trait default values
    
    def _metadata_default(self):
        return FileMetadata(uri="")
    
    def _undo_stack_default(self):
        return UndoStack()
    
    def _bytes_default(self):
        return ""
    
    def _style_default(self):
        return np.zeros(len(self), dtype=np.uint8)
    
    def _segments_default(self):
        r = SegmentData(self.bytes,self.style)
        return list([DefaultSegment(r, 0)])

    #### trait property getters

    def _get_name(self):
        return self.metadata.name or 'Untitled'

    def _get_uri(self):
        return self.metadata.uri

    def _set_uri(self, uri):
        self.metadata.uri = uri

    def _get_read_only(self):
        return self.metadata.read_only

    def _set_read_only(self, read_only):
        self.metadata.read_only = read_only
    
    def _get_can_revert(self):
        return self.metadata.uri != ""
    
    @property
    def menu_name(self):
        if self.uri:
            return "%s (%s)" % (self.name, self.uri)
        return self.name
    
    @classmethod
    def get_blank(cls):
        return cls(bytes="")
    
    def __str__(self):
        return "Document(id=%s): %s" % (self.document_id, self.metadata.uri)
    
    def __len__(self):
        return np.alen(self.bytes)
    
    def __getitem__(self, val):
        return self.bytes[val]
    
    @property
    def dirty(self):
        return self.undo_stack.is_dirty()

    def to_bytes(self):
        return self.bytes.tostring()

    def load_permute(self, editor):
        if self.permute:
            self.permute.load(self, editor)

    def filesystem_path(self):
        try:
            fs_, relpath = fs.opener.opener.parse(self.uri)
            if fs_.hassyspath(relpath):
                return fs_.getsyspath(relpath)
        except fs.errors.FSError:
            pass
        return None
    
    @property
    def bytestream(self):
        return StringIO.StringIO(self.bytes)

    def parse_segments(self, parser_list):
        parser_list.append(DefaultSegmentParser)
        for parser in parser_list:
            try:
                r = SegmentData(self.bytes, self.style)
                s = parser(r)
                break
            except InvalidSegmentParser:
                pass
        self.set_segments(s)
    
    def set_segments(self, parser):
        self.segment_parser = parser
        self.segments = []
        self.segments.extend(parser.segments)
        self.segments.extend(self.user_segments)
    
    @property
    def global_segment(self):
        return self.segments[0]
    
    def add_user_segment(self, segment, replace=False):
        if replace:
            if self.find_matching_user_segment(segment):
                return
        self.user_segments.append(segment)
        self.segments.append(segment)
    
    def is_user_segment(self, segment):
        return segment in self.user_segments
    
    def delete_user_segment(self, segment):
        self.user_segments.remove(segment)
        self.segments.remove(segment)

    def find_matching_user_segment(self, segment):
        for s in self.user_segments:
            if len(s) == len(segment) and s.start_addr == segment.start_addr and s.name == segment.name:
                return True
        return False
    
    def find_segment_index(self, segment):
        try:
            return self.segments.index(segment)
        except ValueError:
            return -1
    
    def find_segment_index_by_name(self, name):
        for i, s in enumerate(self.segments):
            if s.name == name:
                return i
        return -1
    
    def find_segment_in_range(self, addr):
        """Assuming segments had a start_addr param, find first segment that
        has addr as a valid address
        """
        for i, s in enumerate(self.segments):
            if addr >= s.start_addr and addr < (s.start_addr + len(s)):
                return i, s, addr - s.start_addr
        return -1, None, None

    def set_emulator(self, emu):
        self.emulator = emu
        self.emulator_change_event = True
    
    def init_baseline(self, metadata, bytes):
        d = Document(metadata=metadata, bytes=bytes)
        d.parse_segments([])
        self.baseline_document = d
    
    def del_baseline(self):
        self.baseline_document = None
    
    def update_baseline(self):
        if self.baseline_document is not None:
            self.change_count += 1
            self.global_segment.compare_segment(self.baseline_document.global_segment)
    
    def clear_baseline(self):
        self.change_count += 1
        self.global_segment.clear_style_bits(diff=True)
    
    @property
    def has_baseline(self):
        return self.baseline_document is not None




from traits.api import HasTraits, provides

from atrcopy import iter_parsers, SegmentData

from omnivore.file_type.i_file_recognizer import IFileRecognizer
from omnivore.framework.document import Document


@provides(IFileRecognizer)
class AtrcopyRecognizer(HasTraits):
    name = "Atrcopy Disk Image"
    
    id = "application/vnd.atrcopy"
    
    def identify(self, guess):
        r = SegmentData(guess.numpy)
        mime, parser = iter_parsers(r)
        if parser is not None:
            guess.parser = parser
            return mime
    
    def load(self, guess):
        doc = Document(metadata=guess.metadata, bytes=guess.numpy)
        doc.set_segments(guess.parser)
        from omnivore.utils.extra_metadata import check_builtin
        check_builtin(doc)
        return doc



import re


class BaseSearcher(object):
    def __init__(self, editor, search_text, **kwargs):
        self.search_text = self.get_search_text(search_text)
        if len(self.search_text) > 0:
            self.matches = self.get_matches(editor)
            self.set_style(editor)
        else:
            self.matches = []
    
    def get_search_text(self, text):
        return bytearray(text, "utf-8")
    
    def get_matches(self, editor):
        text = editor.segment.search_copy
        rs = re.escape(str(self.search_text))
        matches = [(i.start(), i.end()) for i in re.finditer(rs, text)]
        return matches
    
    def set_style(self, editor):
        editor.segment.set_style_ranges(self.matches, match=True)

class HexSearcher(BaseSearcher):
    def __str__(self):
        return "hex matches: %s" % str(self.matches)
    
    def get_search_text(self, text):
        return bytearray.fromhex(text)

class CharSearcher(BaseSearcher):
    def __init__(self, editor, search_text, match_case=False, find_inverse=True, **kwargs):
        self.match_case = match_case
        self.find_inverse = find_inverse
        BaseSearcher.__init__(self, editor, search_text)
    
    def __str__(self):
        return "char matches: %s" % str(self.matches)

class DisassemblySearcher(BaseSearcher):
    def __init__(self, editor, search_text, match_case=False, find_inverse=True, **kwargs):
        self.match_case = match_case
        self.find_inverse = find_inverse
        BaseSearcher.__init__(self, editor, search_text)
    
    def __str__(self):
        return "disasm matches: %s" % str(self.matches)
    
    def get_search_text(self, text):
        return text
    
    def get_matches(self, editor):
        matches = editor.disassembly.search(self.search_text, self.match_case)
        return matches

known_searchers = [
    HexSearcher,
    CharSearcher,
    DisassemblySearcher,
    ]



#-------------------------------------------------------------------------------
#
#  Copyright (c) 2007, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   03/11/2007
#
#-------------------------------------------------------------------------------

""" Traits UI MS Internet Explorer editor.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import re
import webbrowser

import wx

if wx.Platform == '__WXMSW__':
    # The new version of IEHTMLWindow (wx 2.8.8.0) is mostly compatible with
    # the old one, but it has changed the API for handling COM events, so we
    # cannot use it.
    try:
        import wx.lib.iewin_old as iewin
    except ImportError:
        import wx.lib.iewin as iewin

from traits.api \
    import Bool, Event, Property, Str

from traitsui.wx.editor \
    import Editor

from traitsui.basic_editor_factory \
    import BasicEditorFactory

#-------------------------------------------------------------------------------
#  Constants
#-------------------------------------------------------------------------------

RELATIVE_OBJECTS_PATTERN = re.compile(r'src=["\'](?!https?:)([\s\w/\.]+?)["\']',
                                      re.IGNORECASE)

#-------------------------------------------------------------------------------
#  '_IEHTMLEditor' class:
#-------------------------------------------------------------------------------

class _IEHTMLEditor ( Editor ):
    """ Traits UI MS Internet Explorer editor.
    """

    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    # Is the table editor is scrollable? This value overrides the default.
    scrollable = True

    # External objects referenced in the HTML are relative to this url
    base_url = Str

    # Event fired when the browser home page should be displayed:
    home = Event

    # Event fired when the browser should show the previous page:
    back = Event

    # Event fired when the browser should show the next page:
    forward = Event

    # Event fired when the browser should stop loading the current page:
    stop = Event

    # Event fired when the browser should refresh the current page:
    refresh = Event

    # Event fired when the browser should search the current page:
    search = Event

    # The current browser status:
    status = Str

    # The current browser page title:
    title = Str

    # The URL of the page that just finished loading:
    page_loaded = Str

    # The current page content as HTML:
    html = Property

    #---------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #---------------------------------------------------------------------------

    def init ( self, parent ):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.control = ie = iewin.IEHtmlWindow( parent, -1,
                                      style = wx.NO_FULL_REPAINT_ON_RESIZE )
        self.set_tooltip()

        factory = self.factory
        self.base_url = factory.base_url
        self.sync_value( factory.home,          'home',        'from' )
        self.sync_value( factory.back,          'back',        'from' )
        self.sync_value( factory.forward,       'forward',     'from' )
        self.sync_value( factory.stop,          'stop',        'from' )
        self.sync_value( factory.refresh,       'refresh',     'from' )
        self.sync_value( factory.search,        'search',      'from' )
        self.sync_value( factory.status,        'status',      'to' )
        self.sync_value( factory.title,         'title',       'to' )
        self.sync_value( factory.page_loaded,   'page_loaded', 'to' )
        self.sync_value( factory.html,          'html',        'to' )
        self.sync_value( factory.base_url_name, 'base_url',    'from' )

        parent.Bind( iewin.EVT_StatusTextChange, self._status_modified,     ie )
        parent.Bind( iewin.EVT_TitleChange,      self._title_modified,      ie )
        parent.Bind( iewin.EVT_DocumentComplete, self._page_loaded_modified,ie )
        parent.Bind( iewin.EVT_NewWindow2,       self._new_window_modified, ie )
        parent.Bind( iewin.EVT_BeforeNavigate2,  self._navigate_requested,  ie )

    #---------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #---------------------------------------------------------------------------

    def update_editor ( self ):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        value = self.str_value.strip()

        # We can correct URLs via the BeforeNavigate Event, but the COM
        # interface provides no such option for images. Sadly, we are forced
        # to take a more brute force approach.
        if self.base_url:
            rep = lambda m: r'src="%s%s"' % ( self.base_url, m.group( 1 ) )
            value = re.sub( RELATIVE_OBJECTS_PATTERN, rep, value )

        if value == '':
            self.control.LoadString( '<html><body></body></html>' )

        elif value[:1] == '<':
            self.control.LoadString( value )

        elif (value[:4] != 'http') or (value.find( '://' ) < 0):
            try:
                file = open( value, 'rb' )
                self.control.LoadStream( file )
                file.close()
            except:
                pass

        else:
            self.control.Navigate( value )

    #-- Property Implementations -----------------------------------------------

    def _get_html ( self ):
        return self.control.GetText()

    def _set_html ( self, value ):
        self.control.LoadString( value )

    #-- Event Handlers ---------------------------------------------------------

    def _home_changed ( self ):
        self.control.GoHome()

    def _back_changed ( self ):
        self.control.GoBack()

    def _forward_changed ( self ):
        self.control.GoForward()

    def _stop_changed ( self ):
        self.control.Stop()

    def _search_changed ( self ):
        self.control.GoSearch()

    def _refresh_changed ( self ):
        self.control.Refresh( iewin.REFRESH_COMPLETELY )

    def _status_modified ( self, event ):
        self.status = event.Text

    def _title_modified ( self, event ):
        self.title = event.Text

    def _page_loaded_modified ( self, event ):
        self.page_loaded = event.URL
        self.trait_property_changed( 'html', '', self.html )

    def _new_window_modified ( self, event ):
        # If the event is cancelled, new windows can be disabled.
        # At this point we've opted to allow new windows
        pass

    def _navigate_requested ( self, event ):
        # The way NavigateToString works is to navigate to about:blank then
        # load the supplied HTML into the document property. This borks
        # relative URLs.
        if event.URL.startswith ( 'about:' ):
            base = self.base_url
            if not base.endswith( '/' ):
                base += '/'
            event.URL = base + event.URL[6:]

        if self.factory.open_externally:
            event.Cancel = True
            webbrowser.get( 'windows-default' ).open_new( event.URL )

#-------------------------------------------------------------------------------
#  Create the editor factory object:
#-------------------------------------------------------------------------------

# wxPython editor factory for MS Internet Explorer editors:
class IEHTMLEditor ( BasicEditorFactory ):

    # The editor class to be created:
    klass = _IEHTMLEditor

    # External objects referenced in the HTML are relative to this url
    base_url = Str

    # The object trait containing the base URL
    base_url_name = Str

    # Should links be opened in an external browser?
    open_externally = Bool(False)

    # Optional name of trait used to tell browser to show Home page:
    home = Str

    # Optional name of trait used to tell browser to view the previous page:
    back = Str

    # Optional name of trait used to tell browser to view the next page:
    forward = Str

    # Optional name of trait used to tell browser to stop loading page:
    stop = Str

    # Optional name of trait used to tell browser to refresh the current page:
    refresh = Str

    # Optional name of trait used to tell browser to search the current page:
    search = Str

    # Optional name of trait used to contain the current browser status:
    status = Str

    # Optional name of trait used to contain the current browser page title:
    title = Str

    # Optional name of trait used to contain the URL of the page that just
    # completed loading:
    page_loaded = Str

    # Optional name of trait used to get/set the page content as HTML:
    html = Str




""" Text editor sample task

"""
# Enthought library imports.
from pyface.action.api import Separator, Group
from pyface.tasks.api import Task, TaskWindow, TaskLayout, PaneItem, IEditor, \
    IEditorAreaPane, EditorAreaPane, Editor, DockPane, HSplitter, VSplitter
from pyface.tasks.action.api import SMenuBar, SMenu, SToolBar, SchemaAddition
from traits.api import on_trait_change, Property, Instance, Any, Event, Int

from omnivore.framework.task import FrameworkTask
from omnivore.framework.actions import *
from hex_editor import HexEditor
from preferences import HexEditPreferences
from actions import *
import pane_layout
import omnivore.arch.fonts as fonts
import omnivore.arch.colors as colors
import omnivore.arch.machine as machine
from omnivore.utils.segmentutil import iter_known_segment_parsers
from grid_control import ByteTable
from disassembly import DisassemblyTable


class HexEditTask(FrameworkTask):
    """ Binary file editor
    """

    new_file_text = "Binary File"
    
    hex_grid_lower_case = Bool(True)
    
    assembly_lower_case = Bool(False)

    #### Task interface #######################################################

    id = pane_layout.task_id_with_pane_layout
    name = 'Hex Editor'
    
    preferences_helper = HexEditPreferences
    
    #### Menu events ##########################################################
    
    machine_menu_changed = Event

    emulator_changed = Event
    
    segments_changed = Event
    
    # Must use different trait event in order for actions populated in the
    # dynamic menu (set by segments_changed event above) to have their radio
    # buttons updated properly
    segment_selected = Event

    ###########################################################################
    # 'Task' interface.
    ###########################################################################
    
    def _hex_grid_lower_case_default(self):
        prefs = self.get_preferences()
        return prefs.hex_grid_lower_case
    
    def _assembly_lower_case_default(self):
        prefs = self.get_preferences()
        return prefs.assembly_lower_case

    def _default_layout_default(self):
        return pane_layout.pane_layout()

    def create_dock_panes(self):
        return pane_layout.pane_create()

    def _extra_actions_default(self):
        data_menu = self.create_menu("Menu", "Disk Image", "ParserGroup", "EmulatorGroup", "ActionGroup")
        segment_menu = self.create_menu("Menu", "Segments", "SegmentGroup")
        bytes_menu = self.create_menu("Menu", "Bytes", "HexModifyGroup")
        actions = [
            # Menubar additions
            SchemaAddition(factory=lambda: bytes_menu,
                           path='MenuBar',
                           after="Edit",
                           ),
            SchemaAddition(factory=lambda: segment_menu,
                           path='MenuBar',
                           after="Edit",
                           ),
            SchemaAddition(factory=lambda: data_menu,
                           path='MenuBar',
                           after="Edit",
                           ),
            ]
        return actions

    def _active_editor_changed(self, editor):
        # Make sure it's a valid document before refreshing
        if editor is not None and editor.document.segments:
            editor.update_panes()

    ###########################################################################
    # 'FrameworkTask' interface.
    ###########################################################################
    
    def initialize_class_preferences(self):
        prefs = self.get_preferences()
        ByteTable.update_preferences(prefs)
        DisassemblyTable.update_preferences(prefs)

    def get_editor(self, guess=None):
        """ Opens a new empty window
        """
        editor = HexEditor()
        return editor

    @on_trait_change('window.application.preferences_changed_event')
    def refresh_from_new_preferences(self):
        e = self.active_editor
        if e is not None:
            prefs = self.get_preferences()
            e.machine.text_font = prefs.text_font
            e.map_width = prefs.map_width
            e.bitmap_width = prefs.bitmap_width
            self.hex_grid_lower_case = prefs.hex_grid_lower_case
            self.assembly_lower_case = prefs.assembly_lower_case
            ByteTable.update_preferences(prefs)
            DisassemblyTable.update_preferences(prefs)
            e.reconfigure_panes()

    def get_font_mapping_actions(self, task=None):
        # When actions are needed for a popup, the task must be supplied. When
        # used from a menubar, supplying the task can mess up the EditorAction
        # automatic trait stuff, so this hack is needed. At least it is now,
        # after the addition of the Machine menu in b17fa9fe9. For some reason.
        if task is not None:
            kwargs = {'task': task}
        else:
            kwargs = {}
        actions = []
        for m in machine.predefined['font_mapping']:
            actions.append(FontMappingAction(font_mapping=m, **kwargs))
        return actions
    
    def get_actions_Menu_File_ImportGroup(self):
        return [
            InsertFileAction(),
            ]
    
    def get_actions_Menu_File_ExportGroup(self):
        return [
            SaveAsXEXAction(),
            SaveAsXEXBootAction(),
            ]
    
    def get_actions_Menu_File_SaveGroup(self):
        return [
            SaveAction(),
            SaveAsAction(),
            SMenu(SaveSegmentGroup(),
                  id='SaveSegmentAsSubmenu', name="Save Segment As"),
            ]

    def get_actions_Menu_Edit_UndoGroup(self):
        return [
            UndoAction(),
            RedoAction(),
            RevertToBaselineAction(),
            ]
    
    def get_actions_Menu_Edit_CopyPasteGroup(self):
        return [
            CutAction(),
            CopyAction(),
            CopyDisassemblyAction(),
            CopyAsReprAction(),
            PasteAction(),
            PasteAndRepeatAction(),
            ]
    
    def get_actions_Menu_Edit_SelectGroup(self):
        return [
            SelectAllAction(),
            SelectNoneAction(),
            SelectInvertAction(),
            SMenu(
                MarkSelectionAsCodeAction(name="Code"),
                MarkSelectionAsDataAction(name="Data"),
                MarkSelectionAsDisplayListAction(name="ANTIC Display List"),
                MarkSelectionAsJumpmanLevelAction(name="Jumpman Level Description"),
                MarkSelectionAsJumpmanHarvestAction(name="Jumpman Harvest Table"),
                id="mark1", name="Mark Selection As"),
            ]
    
    def get_actions_Menu_Edit_FindGroup(self):
        return [
            FindAction(),
            FindAlgorithmAction(),
            FindNextAction(),
            FindToSelectionAction(),
            ]
    
    def get_actions_Menu_View_ViewConfigGroup(self):
        return [
            ViewDiffHighlightAction(),
            TextFontAction(),
            ]
    
    def get_predefined_machines_actions(self):
        actions = []
        for m in machine.predefined['machine']:
            actions.append(PredefinedMachineAction(machine=m))
        return actions
    
    def get_actions_Menu_View_ViewPredefinedGroup(self):
        machines = self.get_predefined_machines_actions()
        return [
            SMenu(
                Group(
                    *machines,
                    id="a1", separator=True),
                id='MachineChoiceSubmenu1', separator=False, name="Predefined Machines"),
            ]
    
    def get_font_renderer_actions(self):
        actions = []
        for r in machine.predefined['font_renderer']:
            actions.append(FontRendererAction(font_renderer=r))
        return actions
    
    def get_bitmap_renderer_actions(self):
        actions = []
        for r in machine.predefined['bitmap_renderer']:
            actions.append(BitmapRendererAction(bitmap_renderer=r))
        return actions
    
    def get_processor_type_actions(self):
        actions = []
        for r in machine.predefined['disassembler']:
            actions.append(ProcessorTypeAction(disassembler=r))
        return actions
    
    def get_memory_map_actions(self):
        actions = []
        for r in machine.predefined['memory_map']:
            actions.append(MemoryMapAction(memory_map=r))
        return actions
    
    def get_actions_Menu_View_ViewChangeGroup(self):
        font_mapping_actions = self.get_font_mapping_actions()
        font_renderer_actions = self.get_font_renderer_actions()
        bitmap_renderer_actions = self.get_bitmap_renderer_actions()
        processor_type_actions = self.get_processor_type_actions()
        memory_map_actions = self.get_memory_map_actions()
        return [
            SMenu(
                Group(
                    *processor_type_actions,
                    id="a1", separator=True),
                id='mm1', separator=True, name="Processor"),
            SMenu(
                AssemblerChoiceGroup(id="a2", separator=True),
                Group(
                    AddNewAssemblerAction(),
                    EditAssemblersAction(),
                    SetSystemDefaultAssemblerAction(),
                    id="a3", separator=True),
                id='mm2', separator=False, name="Assembler Syntax"),
            SMenu(
                Group(
                    *memory_map_actions,
                    id="a1", separator=True),
                id='mm3', separator=False, name="Memory Map"),
            SMenu(
                Group(
                    ColorStandardAction(name="NTSC", color_standard=0),
                    ColorStandardAction(name="PAL", color_standard=1),
                    id="a0", separator=True),
                Group(
                    UseColorsAction(name="Powerup Colors", colors=colors.powerup_colors()),
                    id="a1", separator=True),
                Group(
                    AnticColorAction(),
                    id="a2", separator=True),
                id='mm4', separator=False, name="Colors"),
            SMenu(
                Group(
                    UseFontAction(font=fonts.A8DefaultFont),
                    UseFontAction(font=fonts.A8ComputerFont),
                    UseFontAction(font=fonts.A2DefaultFont),
                    UseFontAction(font=fonts.A2MouseTextFont),
                    id="a1", separator=True),
                FontChoiceGroup(id="a2", separator=True),
                Group(
                    LoadFontAction(),
                    GetFontFromSelectionAction(),
                    id="a3", separator=True),
                id='mm5', separator=False, name="Font"),
            SMenu(
                Group(
                    *font_renderer_actions,
                    id="a1", separator=True),
                Group(
                    *font_mapping_actions,
                    id="a2", separator=True),
                Group(
                    FontMappingWidthAction(),
                    id="a3", separator=True),
                id='mm6', separator=False, name="Character Display"),
            SMenu(
                Group(
                    *bitmap_renderer_actions,
                    id="a1", separator=True),
                Group(
                    BitmapWidthAction(),
                    BitmapZoomAction(),
                    id="a1", separator=True),
                id='mm7', separator=False, name="Bitmap Display"),
            ]
    
    def get_actions_Menu_DiskImage_EmulatorGroup(self):
        return [
            RunEmulatorAction(id="a3"),
            SMenu(
                EmulatorChoiceGroup(id="a2"),
                Group(
                    AddNewEmulatorAction(),
                    EditEmulatorsAction(),
                    SetSystemDefaultEmulatorAction(),
                    id="a3", separator=True),
                id='MachineEmulator1', name="Emulators"),
            ]
    
    def get_actions_Menu_DiskImage_ParserGroup(self):
        groups = []
        for mime, pretty, parsers in iter_known_segment_parsers():
            actions = [SegmentParserAction(segment_parser=s) for s in parsers]
            if not pretty:
                groups.append(Group(CurrentSegmentParserAction(), separator=True))
                groups.append(Group(*actions, separator=True))
            else:
                groups.append(SMenu(Group(*actions, separator=True), name=pretty))
        return [
            SMenu(
                *groups,
                id='submenu1', separator=False, name="File Type"),
            ]
    
    def get_actions_Menu_DiskImage_ActionGroup(self):
        return [
            GetSegmentFromSelectionAction(),
            MultipleSegmentsFromSelectionAction(),
            InterleaveSegmentsAction(),
            SetSegmentOriginAction(),
            Separator(),
            AddCommentAction(),
            RemoveCommentAction(),
            Separator(),
            LoadBaselineVersionAction(),
            FindNextBaselineDiffAction(),
            FindPrevBaselineDiffAction(),
            ListDiffAction(),
            Separator(),
            SegmentGotoAction(),
            ]
    
    def get_actions_Menu_Segments_SegmentGroup(self):
        return [
            SegmentChoiceGroup(id="a2", separator=True),
            ]
    
    def get_actions_Menu_Bytes_HexModifyGroup(self):
        return [
            ZeroAction(),
            FFAction(),
            NOPAction(),
            SetValueAction(),
            Separator(),
            SetHighBitAction(),
            ClearHighBitAction(),
            BitwiseNotAction(),
            OrWithAction(),
            AndWithAction(),
            XorWithAction(),
            Separator(),
            LeftShiftAction(),
            RightShiftAction(),
            LeftRotateAction(),
            RightRotateAction(),
            ReverseBitsAction(),
            Separator(),
            AddValueAction(),
            SubtractValueAction(),
            SubtractFromAction(),
            MultiplyAction(),
            DivideByAction(),
            DivideFromAction(),
            Separator(),
            RampUpAction(),
            RampDownAction(),
            ]

    def get_keyboard_actions(self):
        return [
            FindPrevAction(),
            CancelMinibufferAction(),
            UndoCursorPositionAction(),
            RedoCursorPositionAction(),
            ]

    ###
    @classmethod
    def can_edit(cls, document):
        return document.metadata.mime == "application/octet-stream" or document.segments
    
    @classmethod
    def get_match_score(cls, document):
        """Return a number based on how good of a match this task is to the
        incoming Document.
        
        0 = generic match
        ...
        10 = absolute match
        """
        return 1



#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms described in the PyQt GPL exception also apply

#
# Author: Riverbank Computing Limited
#------------------------------------------------------------------------------

""" Defines various directory editor for the PyQt user interface toolkit.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

from pyface.qt import QtGui

# FIXME: ToolkitEditorFactory is a proxy class defined here just for backward
# compatibility. The class has been moved to the
# traitsui.editors.custom_editor file.
from traitsui.editors.directory_editor \
    import ToolkitEditorFactory

from file_editor \
    import SimpleEditor as SimpleFileEditor, \
           CustomEditor as CustomFileEditor

#-------------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------------

class SimpleEditor ( SimpleFileEditor ):
    """ Simple style of editor for directories, which displays a text field
        and a **Browse** button that opens a directory-selection dialog box.
    """

    #---------------------------------------------------------------------------
    #  Creates the correct type of file dialog:
    #---------------------------------------------------------------------------

    def _create_file_dialog ( self ):
        """ Creates the correct type of file dialog.
        """
        dlg = QtGui.QFileDialog(self.control)
        dlg.selectFile(self._file_name.text())
        dlg.setFileMode(QtGui.QFileDialog.Directory)
        dlg.setOptions(QtGui.QFileDialog.ShowDirsOnly)

        return dlg

#-------------------------------------------------------------------------------
#  'CustomEditor' class:
#-------------------------------------------------------------------------------

class CustomEditor ( CustomFileEditor ):
    """ Custom style of editor for directories, which displays a tree view of
        the file system.
    """

    #---------------------------------------------------------------------------
    #  Handles the user changing the contents of the edit control:
    #---------------------------------------------------------------------------

    def update_object(self, idx):
        """ Handles the user changing the contents of the edit control.
        """
        if self._model.isDir(idx):
            self.value = unicode(self._model.filePath(idx))



import sys

import numpy as np
import wx
import wx.grid as Grid

import logging
log = logging.getLogger(__name__)


class FontGridRenderer(Grid.PyGridCellRenderer):
    def __init__(self, table, editor):
        """Render data in the specified color and font and fontsize"""
        Grid.PyGridCellRenderer.__init__(self)
        self.table = table
        self.color = editor.text_color
        self.font = editor.text_font
        self.selected_background = editor.highlight_color
        self.selected_brush = wx.Brush(editor.highlight_color, wx.SOLID)
        self.selected_pen = wx.Pen(editor.highlight_color, 1, wx.SOLID)
        self.normal_background = editor.background_color
        self.normal_brush = wx.Brush(editor.background_color, wx.SOLID)
        self.normal_pen = wx.Pen(editor.background_color, 1, wx.SOLID)
        self.colSize = None
        self.rowSize = 50

    def Draw(self, grid, attr, dc, rect, row, col, isSelected):
        # Here we draw text in a grid cell using various fonts
        # and colors.  We have to set the clipping region on
        # the grid's DC, otherwise the text will spill over
        # to the next cell
        dc.SetClippingRect(rect)

        # clear the background
        dc.SetBackgroundMode(wx.SOLID)
        
        index, _ = self.table.get_index_range(row, col)
        if not self.table.is_index_valid(index):
            dc.SetBrush(wx.Brush(wx.WHITE, wx.SOLID))
            dc.SetPen(wx.Pen(wx.WHITE, 1, wx.SOLID))
            dc.DrawRectangleRect(rect)
        else:
            dc.SetBrush(self.normal_brush)
            dc.SetPen(self.normal_pen)
            dc.SetTextBackground(self.normal_background)
            dc.DrawRectangleRect(rect)

            text = self.table.GetValue(row, col)
            dc.SetBackgroundMode(wx.SOLID)

            dc.SetTextForeground(self.color)
            dc.SetFont(self.font)
            dc.DrawText(text, rect.x+1, rect.y+1)

        dc.DestroyClippingRegion()


class FontGridTable(Grid.PyGridTableBase):
    def __init__(self, column_size=20):
        Grid.PyGridTableBase.__init__(self)
        
        self.col_size = 20
        self.parse_tile_map([("test", np.arange(0,10, dtype=np.uint8))])
    
    def parse_tile_map(self, tile_map):
        self.tile_map = tile_map
        self._rows = len(tile_map)
        self._cols = 0
        for label, chars in tile_map:
            n = np.alen(chars)
            if n > self._cols:
                self._cols = n
    
    def get_index_range(self, row, col):
        """Get the byte offset from start of file given row, col
        position.
        """
        index = row * self.bytes_per_row + col
        return index, index

    def get_row_col(self, index):
        return divmod(index, self.bytes_per_row)

    def is_index_valid(self, index):
        return index < self._rows * self._cols
    
    def get_col_size(self, c):
        return self.column_sizes[c]
    
    def get_col_type(self, c):
        return "hex"
   
    def GetNumberRows(self):
        return self._rows

    def GetRowLabelValue(self, row):
        return self.tile_map[row][0]

    def GetNumberCols(self):
        return self._cols

    def GetColLabelValue(self, col):
        return ""
    
    def GetValue(self, row, col):
        label, chars = self.tile_map[row]
        if col >= np.alen(chars):
            return ""
        return str(chars[col])

    def SetValue(self, row, col, value):
        raise NotImplementedError

    def ResetViewProcessArgs(self, *args):
        pass

    def ResetView(self, grid, *args):
        """
        (Grid) -> Reset the grid view.   Call this to
        update the grid if rows and columns have been added or deleted
        """
        oldrows=self._rows
        oldcols=self._cols
        self.ResetViewProcessArgs(*args)
        
        grid.BeginBatch()

        for current, new, delmsg, addmsg in [
            (oldrows, self._rows, Grid.GRIDTABLE_NOTIFY_ROWS_DELETED, Grid.GRIDTABLE_NOTIFY_ROWS_APPENDED),
            (oldcols, self._cols, Grid.GRIDTABLE_NOTIFY_COLS_DELETED, Grid.GRIDTABLE_NOTIFY_COLS_APPENDED),
        ]:

            if new < current:
                msg = Grid.GridTableMessage(self,delmsg,new,current-new)
                grid.ProcessTableMessage(msg)
            elif new > current:
                msg = Grid.GridTableMessage(self,addmsg,new-current)
                grid.ProcessTableMessage(msg)
                self.UpdateValues(grid)
        grid.EndBatch()

        # update the scrollbars and the displayed part of the grid
        dc = wx.MemoryDC()
        dc.SetFont(grid.editor.text_font)
        (width, height) = dc.GetTextExtent("M")
        grid.SetDefaultRowSize(height)
        grid.SetColMinimalAcceptableWidth(width)
        grid.SetRowMinimalAcceptableHeight(height + 1)

        for row in range(self._rows):
            # Can't share GridCellAttrs among columns; causes crash when
            # freeing them.  So, have to individually allocate the attrs for
            # each column
            cellattr = Grid.GridCellAttr()
            cellattr.SetFont(grid.editor.text_font)
            cellattr.SetBackgroundColour("white")
            renderer = FontGridRenderer(self, grid.editor)
            cellattr.SetRenderer(renderer)
            log.debug("hexcol %d width=%d" % (col,width))
            grid.SetRowMinimalHeight(col, height)
            grid.SetRowSize(row, (height * 2) + 4)
            grid.SetRowAttr(row, cellattr)

        self._rows = self.GetNumberRows()
        self._cols = self.GetNumberCols()
        
        label_font = grid.editor.text_font.Bold()
        grid.SetLabelFont(label_font)
        dc.SetFont(label_font)
        (width, height) = dc.GetTextExtent("M")
        grid.HideColLabels()
        text = self.GetRowLabelValue(self._rows - 1)
        grid.SetRowLabelSize(width * len(text) + 4)
        
        grid.AdjustScrollbars()
        grid.ForceRefresh()

    def UpdateValues(self, grid):
        """Update all displayed values"""
        # This sends an event to the grid table to update all of the values
        msg = Grid.GridTableMessage(self, Grid.GRIDTABLE_REQUEST_VIEW_GET_VALUES)
        grid.ProcessTableMessage(msg)


class FontGrid(Grid.Grid):
    """
    View for editing in hexidecimal notation.
    """

    def __init__(self, parent, task, table, **kwargs):
        """Create the HexEdit viewer
        """
        Grid.Grid.__init__(self, parent, -1, **kwargs)
        self.task = task
        self.editor = None
        if table is None:
            table = FontGridTable()
        self.table = table

        # The second parameter means that the grid is to take
        # ownership of the table and will destroy it when done.
        # Otherwise you would need to keep a reference to it and call
        # its Destroy method later.
        self.SetTable(self.table, True)
        self.SetMargins(0,0)
        self.SetColMinimalAcceptableWidth(10)
        self.EnableDragGridSize(False)
        self.DisableDragRowSize()

        self.RegisterDataType(Grid.GRID_VALUE_STRING, None, None)

        self.allow_range_select = True
        self.updateUICallback = None
        self.Bind(Grid.EVT_GRID_CELL_LEFT_CLICK, self.OnLeftDown)
        self.GetGridWindow().Bind(wx.EVT_MOTION, self.on_motion)
        self.Bind(Grid.EVT_GRID_CELL_RIGHT_CLICK, self.OnRightDown)
#        self.Bind(Grid.EVT_GRID_SELECT_CELL, self.OnSelectCell)
#        self.Bind(Grid.EVT_GRID_RANGE_SELECT, self.OnSelectRange)
        self.Bind(wx.EVT_KEY_DOWN, self.OnKeyDown)
        self.Show(True)

    def OnRightDown(self, evt):
        log.debug(self.GetSelectedRows())
        actions = self.get_popup_actions(r, c)
        if actions:
            self.editor.popup_context_menu_from_actions(self, actions)
    
    def get_popup_actions(self, r, c):
        return []

    def OnLeftDown(self, evt):
        c, r = (evt.GetCol(), evt.GetRow())
        self.ClearSelection()
        e = self.editor
        e.anchor_initial_start_index, e.anchor_initial_end_index = self.table.get_index_range(r, c)
        e.anchor_start_index, e.anchor_end_index = e.anchor_initial_start_index, e.anchor_initial_end_index
        evt.Skip()
        self.SetGridCursor(r, c)
        wx.CallAfter(self.ForceRefresh)
        wx.CallAfter(self.task.active_editor.index_clicked, e.anchor_start_index, 0, self)
 
    def on_motion(self, evt):
        e = self.editor
        if evt.LeftIsDown():
            x, y = evt.GetPosition()
            x, y = self.CalcUnscrolledPosition(x, y)
            r, c = self.XYToCell(x, y)
            index1, index2 = self.table.get_index_range(r, c)
            update = False
            if e.anchor_start_index <= index1:
                if index2 != e.anchor_end_index:
                    e.anchor_start_index = e.anchor_initial_start_index
                    e.anchor_end_index = index2
                    update = True
            else:
                if index1 != e.anchor_end_index:
                    e.anchor_start_index = e.anchor_initial_end_index
                    e.anchor_end_index = index1
                    update = True
            if update:
                self.SetGridCursor(r, c)
                wx.CallAfter(self.ForceRefresh)
                wx.CallAfter(self.task.active_editor.index_clicked, e.anchor_end_index, 0, self)
        evt.Skip()

    def OnSelectCell(self, evt):
        log.debug("cell selected r=%d c=%d" % (evt.GetCol(), evt.GetRow()))
        evt.Skip()

    def OnKeyDown(self, evt):
        log.debug("evt=%s" % evt)
        key = evt.GetKeyCode()
        moved = False
        r, c = self.GetGridCursorRow(), self.GetGridCursorCol()
        if key == wx.WXK_RETURN or key == wx.WXK_TAB:
            if evt.ControlDown():   # the edit control needs this key
                evt.Skip()
            else:
                self.DisableCellEditControl()
                if evt.ShiftDown():
                    (r, c) = self.table.getPrevCursorPosition(r, c)
                else:
                    (r, c) = self.table.getNextCursorPosition(r, c)
                moved = True
        elif key == wx.WXK_RIGHT:
            r, c = self.table.getNextCursorPosition(r, c)
            moved = True
        elif key == wx.WXK_LEFT:
            r, c = self.table.getPrevCursorPosition(r, c)
            moved = True
        elif key == wx.WXK_UP:
            r = 0 if r <= 1 else r - 1
            moved = True
        elif key == wx.WXK_DOWN:
            n = self.GetNumberRows()
            r = n - 1 if r >= n - 1 else r + 1
            moved = True
        else:
            evt.Skip()
        
        if moved:
            self.SetGridCursor(r, c)
            self.MakeCellVisible(r, c)
            index1, index2 = self.table.get_index_range(r, c)
            wx.CallAfter(self.task.active_editor.index_clicked, index1, 0, self)

    def abortEdit(self):
        self.DisableCellEditControl()

    def goto_index(self, index):
        row, col = self.table.get_row_col(index)
        self.SetGridCursor(row, col)
        self.MakeCellVisible(row,col)

    def select_index(self, cursor):
        self.ClearSelection()
        self.goto_index(cursor)
        self.ForceRefresh()
    
    def change_value(self, row, col, text):
        """Called after editor has provided a new value for a cell.
        
        Can use this to override the default handler.  Return True if the grid
        should be updated, or False if the value is invalid or the grid will
        be updated some other way.
        """
        self.table.SetValue(row, col, text) # update the table
        return True


class TileMapControl(FontGrid):
    """
    View for editing in hexidecimal notation.
    """

    def __init__(self, parent, task, **kwargs):
        """Create the HexEdit viewer
        """
        table = FontGridTable()
        FontGrid.__init__(self, parent, task, table, **kwargs)

    def recalc_view(self):
        editor = self.task.active_editor
        if editor is not None:
            self.editor = editor
            self.table.ResetView(self, editor)
            self.table.UpdateValues(self)
    
    def change_value(self, row, col, text):
        """Called after editor has provided a new value for a cell.
        
        Can use this to override the default handler.  Return True if the grid
        should be updated, or False if the value is invalid or the grid will
        be updated some other way.
        """
        try:
            val = int(text,16)
            if val >= 0 and val < 256:
                start, end = self.table.get_index_range(row, col)
                cmd = ChangeByteCommand(self.table.segment, start, end, val)
                self.task.active_editor.process_command(cmd)
        except ValueError:
            pass
        return False
    
    def get_popup_actions(self, r, c):
        return [CutAction, CopyAction, PasteAction, None, SelectAllAction]



# peppy Copyright (c) 2006-2010 Rob McMullen
# Licenced under the GPLv2; see http://peppy.flipturn.org for more info
import os
import sys
import struct

import wx
import wx.grid as Grid
import wx.lib.newevent

from atrcopy import match_bit_mask, comment_bit_mask, data_bit_mask, selected_bit_mask, diff_bit_mask

from omnivore.utils.wx.bytegrid import ByteGridTable, ByteGrid

from commands import ChangeByteCommand

import logging
log = logging.getLogger(__name__)

# Grid tips:
# http://www.blog.pythonlibrary.org/2010/04/04/wxpython-grid-tips-and-tricks/


class ImageCache(object):
    def __init__(self, width=-1, height=-1):
        self.width = width
        self.height = height
        self.cache = {}
    
    def invalidate(self):
        self.cache = {}
    
    def set_colors(self, editor):
        m = editor.machine
        self.color = m.text_color
        self.diff_color = m.diff_text_color
        self.font = m.text_font
        self.selected_background = m.highlight_color
        self.selected_brush = wx.Brush(m.highlight_color, wx.SOLID)
        self.selected_pen = wx.Pen(m.highlight_color, 1, wx.SOLID)
        self.normal_background = m.background_color
        self.normal_brush = wx.Brush(m.background_color, wx.SOLID)
        self.normal_pen = wx.Pen(m.background_color, 1, wx.SOLID)
        self.data_background = m.data_color
        self.data_brush = wx.Brush(m.data_color, wx.SOLID)
        self.cursor_background = m.background_color
        self.cursor_brush = wx.Brush(m.background_color, wx.TRANSPARENT)
        self.cursor_pen = wx.Pen(m.unfocused_cursor_color, 2, wx.SOLID)
        self.match_background = m.match_background_color
        self.match_brush = wx.Brush(m.match_background_color, wx.SOLID)
        self.match_pen = wx.Pen(m.match_background_color, 1, wx.SOLID)
        self.comment_background = m.comment_background_color
        self.comment_brush = wx.Brush(m.comment_background_color, wx.SOLID)
        self.comment_pen = wx.Pen(m.comment_background_color, 1, wx.SOLID)
    
    def draw_blank(self, dc, rect):
        dc.SetBrush(wx.Brush(wx.WHITE, wx.SOLID))
        dc.SetPen(wx.Pen(wx.WHITE, 1, wx.SOLID))
        dc.DrawRectangleRect(rect)
    
    def draw_text(self, dc, rect, text, style):
        k = (text, style, rect.width, rect.height)
        try:
            bmp = self.cache[k]
        except KeyError:
            bmp = wx.EmptyBitmap(rect.width, rect.height)
            mdc = wx.MemoryDC()
            mdc.SelectObject(bmp)
            r = wx.Rect(0, 0, rect.width, rect.height)
            self.draw_text_to_dc(mdc, r, text, style)
            del mdc  # force the bitmap painting by deleting the gc
            self.cache[k] = bmp
        dc.DrawBitmap(bmp, rect.x, rect.y)
    
    def draw_text_to_dc(self, dc, rect, text, style):
        if style & selected_bit_mask:
            dc.SetBrush(self.selected_brush)
            dc.SetPen(self.selected_pen)
            dc.SetTextBackground(self.selected_background)
        elif style & match_bit_mask:
            dc.SetPen(self.match_pen)
            dc.SetBrush(self.match_brush)
            dc.SetTextBackground(self.match_background)
        elif style & comment_bit_mask:
            dc.SetPen(self.comment_pen)
            dc.SetBrush(self.comment_brush)
            dc.SetTextBackground(self.comment_background)
        elif style & data_bit_mask:
            dc.SetPen(self.normal_pen)
            dc.SetBrush(self.data_brush)
            dc.SetTextBackground(self.data_background)
        else:
            dc.SetPen(self.normal_pen)
            dc.SetBrush(self.normal_brush)
            dc.SetTextBackground(self.normal_background)
        dc.SetBackgroundMode(wx.SOLID)
        dc.DrawRectangleRect(rect)
        if style & diff_bit_mask:
            dc.SetTextForeground(self.diff_color)
        else:
            dc.SetTextForeground(self.color)
        dc.SetFont(self.font)
        dc.DrawText(text, rect.x+1, rect.y+1)


class CachingHexRenderer(Grid.PyGridCellRenderer):
    def __init__(self, table, editor, cache):
        """Render data in the specified color and font and fontsize"""
        Grid.PyGridCellRenderer.__init__(self)
        self.table = table
        self.cache = cache
        cache.set_colors(editor)

    def Draw(self, grid, attr, dc, rect, row, col, isSelected):
        index, _ = self.table.get_index_range(row, col)
        if not self.table.is_index_valid(index):
            self.cache.draw_blank(dc, rect)
        else:
            text, style = self.table.get_value_style(row, col)
            self.cache.draw_text(dc, rect, text, style)

            r, c = self.table.get_row_col(grid.editor.cursor_index)
            if row == r and col == c:
                dc.SetPen(self.cache.cursor_pen)
                dc.SetBrush(self.cache.cursor_brush)
                x = rect.x+1
                if sys.platform == "darwin":
                    w = rect.width - 2
                    h = rect.height - 2
                else:
                    w = rect.width - 1
                    h = rect.height - 1
                dc.DrawRectangle(rect.x+1, rect.y+1, w, h)


class ByteTable(ByteGridTable):
    def __init__(self, bytes_per_row=16):
        ByteGridTable.__init__(self)
        
        self._debug=False
        self.bytes_per_row = bytes_per_row
        self._cols = self.bytes_per_row
        self._rows = 0

    def set_editor(self, editor):
        self.editor = editor
        self.segment = segment = self.editor.segment
        self.start_offset = segment.start_addr & 0x0f
        self._rows=((self.start_offset + len(segment) - 1) / self.bytes_per_row) + 1
        log.debug("segment %s: rows=%d cols=%d len=%d" % (segment, self._rows, self.bytes_per_row, len(segment)))
    
    def get_index_range(self, row, col):
        """Get the byte offset from start of file given row, col
        position.
        """
        index = row * self.bytes_per_row + col - self.start_offset
        return index, index + 1

    def get_row_col(self, index):
        return divmod(index + self.start_offset, self.bytes_per_row)

    def is_index_valid(self, index):
        return 0 <= index < len(self.segment)
    
    def get_col_size(self, col):
        return 2

    def get_value_style_upper(self, row, col):
        i, _ = self.get_index_range(row, col)
        return "%02X" % self.segment[i], self.segment.style[i]

    def get_value_style_lower(self, row, col):
        i, _ = self.get_index_range(row, col)
        return "%02x" % self.segment[i], self.segment.style[i]

    def get_label_at_index(self, index):
        return self.segment.label(index, self.get_value_style == self.get_value_style_lower)

    def GetRowLabelValue(self, row):
        return self.get_label_at_index(row*self.bytes_per_row - self.start_offset)

    def GetColLabelValue(self, col):
        if self.get_value_style == self.get_value_style_lower:
            return "%x" % col
        return "%X" % col

    def SetValue(self, row, col, value):
        val=int(value,16)
        if val>=0 and val<256:
            i, _ = self.get_index_range(row, col)
            if self.is_index_valid(i):
                self.segment[i:i+1] = val
                return True
            log.debug('SetValue(%d, %d, "%s")=%d index %d out of range.' % (row, col, value, val, i))
            return False
        else:
            log.debug('SetValue(%d, %d, "%s")=%d out of range.' % (row, col, value, val))
            return False

    def ResetViewProcessArgs(self, grid, editor, *args):
        grid.image_cache.invalidate()
        self.set_editor(editor)


class HexEditControl(ByteGrid):
    """
    View for editing in hexidecimal notation.
    """

    def __init__(self, parent, task, **kwargs):
        """Create the HexEdit viewer
        """
        table = ByteTable()
        ByteGrid.__init__(self, parent, task, table, **kwargs)
        self.image_cache = ImageCache()
    
    def get_grid_cell_renderer(self, table, editor):
        return CachingHexRenderer(table, editor, self.image_cache)
    
    def get_status_message_at_index(self, index, row, col):
        msg = ByteGrid.get_status_message_at_index(self, index, row, col)
        comments = self.table.segment.get_comment(index)
        return "%s  %s" % (msg, comments)
    
    def change_value(self, row, col, text):
        """Called after editor has provided a new value for a cell.
        
        Can use this to override the default handler.  Return True if the grid
        should be updated, or False if the value is invalid or the grid will
        be updated some other way.
        """
        try:
            val = int(text,16)
            if val >= 0 and val < 256:
                start, end = self.table.get_index_range(row, col)
                if self.table.is_index_valid(start):
                    cmd = ChangeByteCommand(self.table.segment, start, end, val)
                    self.task.active_editor.process_command(cmd)
        except ValueError:
            pass
        return False
    
    def get_popup_actions(self, r, c):
        return self.editor.common_popup_actions()



#!/usr/bin/env python
""" Convert udis module level data into a single dictionary structure for all
known processors and save it into cputables.py

"""
import os
import glob

# flags
pcr = 1
und = 2
z80bit = 4
r = 64
w = 128

def find_nop(cpu):
    table = cpu['opcodeTable']
    possibilities = []
    nop = 0x00
    for opcode, optable in table.items():
        try:
            length, mnemonic, mode, flag = optable
        except ValueError:
            length, mnemonic, mode = optable
            flag = 0
        if mnemonic == "nop" and flag == 0:
            nop = opcode
            break
    return nop


def read_udis(pathname):
    """ Read all the processor-specific opcode info and pull into a container
    dictionary keyed on the processor name.
    
    The udis files have module level data, so this pulls the data from multiple
    cpus into a single structure that can then be refereced by processor name.
    For example, to find the opcode table in the generated dictionary for the
    6502 processor, use:
    
    cpus['6502']['opcodeTable']
    """
    files = glob.glob("%s/*.py" % pathname)
    cpus = {}
    for filename in files:
        localfile = os.path.basename(filename)
        with open(filename, "r") as fh:
            source = fh.read()
            if "import cputables" in source:
                continue
            if "addressModeTable" in source and "opcodeTable" in source:
                cpu_name, _ = os.path.splitext(localfile)
                g = {"pcr": pcr, "und": und, "r": r, "w": w, "z80bit": z80bit}
                d = {}
                try:
                    exec(source, g, d)
                    if 'opcodeTable' in d:
                        cpus[cpu_name] = d
                        nop = find_nop(d)
                        cpus[cpu_name]["nop"] = nop
                except SyntaxError:
                    # ignore any python 3 files
                    pass
    return cpus


if __name__ == "__main__":
    import sys
    import argparse
    
    supported_cpus = read_udis(".")
    output = []
    import pprint
    output.append("# Autogenerated from udis source! Do not edit here, change udis source instead.")
    output.append("processors =\\")
    for line in pprint.pformat(supported_cpus).splitlines():
        output.append(line.strip())
#    print supported_cpus
    with open("cputables.py", "w") as fh:
        fh.write("\n".join(output))
        fh.write("\n")



import sys
import wx

import logging
log = logging.getLogger(__name__)

class TriggerList(wx.ListBox):
    """Trigger selector for choosing which trigger actions to edit
    """

    def __init__(self, parent, task, **kwargs):
        self.task = task
        self.triggers = None
        
        wx.ListBox.__init__(self, parent, style=wx.LB_SINGLE, **kwargs)
        self.Bind(wx.EVT_LEFT_DOWN, self.on_left_down)
        self.Bind(wx.EVT_LISTBOX, self.on_click)
        self.Bind(wx.EVT_LISTBOX_DCLICK, self.on_dclick)
        self.Bind(wx.EVT_MOTION, self.on_tooltip)
    
    def set_task(self, task):
        self.task = task
     
    def recalc_view(self):
        self.editor = e = self.task.active_editor
        if e is not None:
            selected = e.bitmap.trigger_root
            self.set_peanuts(selected)
    
    def refresh_view(self):
        e = self.editor
        if e is not None:
            if self.IsShown():
                log.debug("refreshing %s" % self)
                self.recalc_view()
            else:
                log.debug("skipping refresh of hidden %s" % self)
   
    def parse_peanuts(self, peanuts, items, triggers, indent=""):
        for peanut in peanuts:
            items.append(indent + peanut.trigger_str)
            triggers.append(peanut)
            children = []
            for p in peanut.trigger_painting:
                if p.single:
                    children.append(p)
            if children:
                self.parse_peanuts(children, items, triggers, indent + "    ")

    def set_peanuts(self, selected=None):
        items = ["Main Level"]
        triggers = [None]
        index = 1
        selected_index = 0
        state = self.editor.bitmap.screen_state
        if state is not None:
            self.parse_peanuts(state.sorted_peanuts, items, triggers)
            for index, trigger in enumerate(triggers):
                if trigger == selected:
                    selected_index = index
        if len(items) != self.GetCount():
            self.SetItems(items)
        else:
            for i, item in enumerate(items):
                old = self.GetString(i)
                if old != item:
                    self.SetString(i, item)
        if selected_index < self.GetCount():
            self.SetSelection(selected_index)
        self.triggers = triggers

    def on_left_down(self, event):
        item = self.HitTest(event.GetPosition())
        if item >= 0:
            selected = self.GetSelection()
            if selected != item:
                e = self.editor
                wx.CallAfter(e.set_trigger_view, self.triggers[item])
        event.Skip()

    def on_click(self, event):
        # BUG: doesn't seem to get called when selecting a segment, using the
        # comments sidebar to jump to another segment, then attempting to
        # select that previous segment. This function never gets called in
        # that case, so I had to add the check on EVT_LEFT_DOWN
        is_selected = event.GetExtraLong()
        if is_selected:
            selected = event.GetSelection()
            e = self.editor
            wx.CallAfter(e.set_trigger_view, self.triggers[selected])
        event.Skip()
    
    def on_dclick(self, event):
        event.Skip()
    
    def on_tooltip(self, evt):
        pos = evt.GetPosition()
        selected = self.HitTest(pos)
        if selected >= 0:
            self.task.status_bar.message = "peanut #%d" % selected
        else:
            self.task.status_bar.message = ""
        evt.Skip()



import os

import pytest

import miniasm


cputests = [
    ("65816", "6f67726182f61c82fdff8298e910f882a27182ee7e"),
    ("z80", "0e25dd21aae6dd3655ff384022804cdd39fd39ddcbf0f5ddcb0101ddcb0102ddcb0202ddcb7f02fdcbe2a9ddcbe2a9"),
    ]

class TestMiniasm(object):
    def setup(self):
        self.start_pc = 0xa000

    @pytest.mark.parametrize("cpu,binary", cputests)
    def test_cpu(self, cpu, binary):
        source = binary.decode("hex")
        success, failure = miniasm.process(source, cpu, self.start_pc, cpu)
        assert failure == 0

if __name__ == "__main__":
    t = TestMiniasm()
    t.setup()
    for cpu, binary in cputests:
        t.test_cpu(cpu, binary)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   11/01/2004
#
#------------------------------------------------------------------------------

""" Creates a wxPython user interface for a specified UI object.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import wx

from helper \
    import restore_window, save_window, TraitsUIScrolledPanel

from ui_base \
    import BaseDialog

from ui_panel \
    import panel, show_help

from constants \
    import DefaultTitle, WindowColor, screen_dy, scrollbar_dx

from traitsui.menu \
    import ApplyButton, RevertButton, OKButton, CancelButton, HelpButton

#-------------------------------------------------------------------------------
#  Creates a modal wxPython user interface for a specified UI object:
#-------------------------------------------------------------------------------

def ui_modal ( ui, parent ):
    """ Creates a modal wxPython user interface for a specified UI object.
    """
    ui_dialog( ui, parent, True )

#-------------------------------------------------------------------------------
#  Creates a non-modal wxPython user interface for a specified UI object:
#-------------------------------------------------------------------------------

def ui_nonmodal ( ui, parent ):
    """ Creates a non-modal wxPython user interface for a specified UI object.
    """
    ui_dialog( ui, parent, False )

#-------------------------------------------------------------------------------
#  Creates a wxPython dialog-based user interface for a specified UI object:
#-------------------------------------------------------------------------------

def ui_dialog ( ui, parent, is_modal ):
    """ Creates a wxPython dialog box for a specified UI object.

    Changes are not immediately applied to the underlying object. The user must
    click **Apply** or **OK** to apply changes. The user can revert changes by
    clicking **Revert** or **Cancel**.
    """
    if ui.owner is None:
        ui.owner = ModalDialog()

    ui.owner.init( ui, parent, is_modal )
    ui.control = ui.owner.control
    ui.control._parent = parent
    try:
        ui.prepare_ui()
    except:
        ui.control.Destroy()
        ui.control.ui = None
        ui.control    = None
        ui.owner      = None
        ui.result     = False
        raise

    ui.handler.position( ui.info )
    restore_window( ui )

    if is_modal:
        ui.control.ShowModal()
    else:
        ui.control.Show()

#-------------------------------------------------------------------------------
#  'ModalDialog' class:
#-------------------------------------------------------------------------------

class ModalDialog ( BaseDialog ):
    """ Modal dialog box for Traits-based user interfaces.
    """
    #---------------------------------------------------------------------------
    #  Initializes the object:
    #---------------------------------------------------------------------------

    def init ( self, ui, parent, is_modal ):
        self.is_modal = is_modal
        style         = 0
        view          = ui.view
        if view.resizable:
            style |= wx.RESIZE_BORDER

        title = view.title
        if title == '':
            title = DefaultTitle

        revert = apply = False
        window = ui.control
        if window is not None:
            window.SetSizer( None )
            ui.reset()
            if hasattr( self, 'revert' ):
                revert = self.revert.IsEnabled()
            if hasattr( self, 'apply' ):
                apply = self.apply.IsEnabled()
        else:
            self.ui = ui
            if is_modal:
                window = wx.Dialog( parent, -1, title,
                                    style = style | wx.DEFAULT_DIALOG_STYLE )
            else:
                window = wx.Frame(  parent, -1, title, style = style |
                            (wx.DEFAULT_FRAME_STYLE & (~wx.RESIZE_BORDER) ) )

            window.SetBackgroundColour( WindowColor )
            self.control = window
            self.set_icon( view.icon )
            wx.EVT_CLOSE( window, self._on_close_page )
            wx.EVT_CHAR(  window, self._on_key )

            # Create the 'context' copies we will need while editing:
            context     = ui.context
            ui._context = context
            ui.context  = self._copy_context( context )
            ui._revert  = self._copy_context( context )

        # Create the actual trait sheet panel and imbed it in a scrollable
        # window (if requested):
        sw_sizer = wx.BoxSizer( wx.VERTICAL )
        if ui.scrollable:
            sizer       = wx.BoxSizer( wx.VERTICAL )
            sw          = TraitsUIScrolledPanel( window )
            trait_sheet = panel( ui, sw )
            sizer.Add( trait_sheet, 1, wx.EXPAND | wx.ALL, 4 )
            tsdx, tsdy = trait_sheet.GetSizeTuple()
            tsdx += 8
            tsdy += 8
            sw.SetScrollRate( 16, 16 )
            max_dy = (2 * screen_dy) / 3
            sw.SetSizer( sizer )
            sw.SetSize( wx.Size( tsdx + ((tsdy > max_dy) * scrollbar_dx),
                                 min( tsdy, max_dy ) ) )
        else:
            sw = panel( ui, window )

        sw_sizer.Add( sw, 1, wx.EXPAND )

        buttons  = [ self.coerce_button( button ) for button in view.buttons ]
        nbuttons = len( buttons )
        if (nbuttons != 1) or (not self.is_button( buttons[0], '' )):

            # Create the necessary special function buttons:
            sw_sizer.Add( wx.StaticLine( window, -1 ), 0, wx.EXPAND )
            b_sizer = wx.BoxSizer( wx.HORIZONTAL )

            if nbuttons == 0:
                if view.apply:
                    self.check_button( buttons, ApplyButton )
                    if view.revert:
                        self.check_button( buttons, RevertButton )

                if view.ok:
                    self.check_button( buttons, OKButton )

                if view.cancel:
                    self.check_button( buttons, CancelButton )

                if view.help:
                    self.check_button( buttons, HelpButton )

            for raw_button, button in zip( view.buttons, buttons ):
                default = raw_button == view.default_button
                
                if self.is_button( button, 'Apply' ):
                    self.apply = self.add_button( button, b_sizer,
                            self._on_apply, apply, default = default )
                    ui.on_trait_change( self._on_applyable, 'modified',
                                        dispatch = 'ui' )

                elif self.is_button( button, 'Revert' ):
                    self.revert = self.add_button( button, b_sizer,
                            self._on_revert, revert, default = default )

                elif self.is_button( button, 'OK' ):
                    self.ok = self.add_button( button, b_sizer, self._on_ok,
                                               default = default )
                    ui.on_trait_change( self._on_error, 'errors',
                                        dispatch = 'ui' )

                elif self.is_button( button, 'Cancel' ):
                    self.add_button( button, b_sizer, self._on_cancel,
                                     default = default)

                elif self.is_button( button, 'Help' ):
                    self.add_button( button, b_sizer, self._on_help,
                                     default = default)

                elif not self.is_button( button, '' ):
                    self.add_button( button, b_sizer, default = default )

            sw_sizer.Add( b_sizer, 0, wx.ALIGN_RIGHT | wx.ALL, 5 )

        # Add the menu bar, tool bar and status bar (if any):
        self.add_menubar()
        self.add_toolbar()
        self.add_statusbar()

        # Lay all of the dialog contents out:
        window.SetSizerAndFit( sw_sizer )

    #---------------------------------------------------------------------------
    #  Closes the dialog window:
    #---------------------------------------------------------------------------

    def close ( self, rc = wx.ID_OK ):
        """ Closes the dialog window.
        """
        ui = self.ui
        ui.result = (rc == wx.ID_OK)
        save_window( ui )
        if self.is_modal:
            self.control.EndModal( rc )

        ui.finish()
        self.ui = self.apply = self.revert = self.help = self.control = None

    #---------------------------------------------------------------------------
    #  Creates a copy of a 'context' dictionary:
    #---------------------------------------------------------------------------

    def _copy_context ( self, context ):
        """ Creates a copy of a *context* dictionary.
        """
        result = {}
        for name, value in context.items():
            if value is not None:
                result[ name ] = value.clone_traits()
            else:
                result[ name ] = None

        return result

    #---------------------------------------------------------------------------
    #  Applies the traits in the 'from' context to the 'to' context:
    #---------------------------------------------------------------------------

    def _apply_context ( self, from_context, to_context ):
        """ Applies the traits in the *from_context* to the *to_context*.
        """
        for name, value in from_context.items():
            if value is not None:
                to_context[ name ].copy_traits( value )
            else:
                to_context[ name ] = None

        if to_context is self.ui._context:
            on_apply = self.ui.view.on_apply
            if on_apply is not None:
                on_apply()

    #---------------------------------------------------------------------------
    #  Handles the user clicking the window/dialog 'close' button/icon:
    #---------------------------------------------------------------------------

    def _on_close_page ( self, event ):
        """ Handles the user clicking the window/dialog "close" button/icon.
        """
        if self.ui.view.close_result == True:
            self._on_ok( event )
        else:
            self._on_cancel( event )

    #---------------------------------------------------------------------------
    #  Closes the window and saves changes (if allowed by the handler):
    #---------------------------------------------------------------------------

    def _on_ok ( self, event = None ):
        """ Closes the window and saves changes (if allowed by the handler).
        """
        if self.ui.handler.close( self.ui.info, True ):
            self._apply_context( self.ui.context, self.ui._context )
            self.close( wx.ID_OK )

    #---------------------------------------------------------------------------
    #  Closes the window and discards changes (if allowed by the handler):
    #---------------------------------------------------------------------------

    def _on_cancel ( self, event = None ):
        """ Closes the window and discards changes (if allowed by the handler).
        """
        if self.ui.handler.close( self.ui.info, False ):
            self._apply_context( self.ui._revert, self.ui._context )
            self.close( wx.ID_CANCEL )

    #---------------------------------------------------------------------------
    #  Handles the 'Help' button being clicked:
    #---------------------------------------------------------------------------

    def _on_help ( self, event ):
        """ Handles the **Help** button being clicked.
        """
        self.ui.handler.show_help( self.ui.info, event.GetEventObject() )

    #---------------------------------------------------------------------------
    #  Handles the user hitting the 'Esc'ape key:
    #---------------------------------------------------------------------------

    def _on_key ( self, event ):
        """ Handles the user pressing the Escape key.
        """
        if event.GetKeyCode() == 0x1B:
            self._on_close_page( event )

    #---------------------------------------------------------------------------
    #  Handles an 'Apply' all changes request:
    #---------------------------------------------------------------------------

    def _on_apply ( self, event ):
        """ Handles a request to apply changes.
        """
        ui = self.ui
        self._apply_context( ui.context, ui._context )
        if hasattr(self, 'revert'):
            self.revert.Enable( True )
        ui.handler.apply( ui.info )
        ui.modified = False

    #---------------------------------------------------------------------------
    #  Handles a 'Revert' all changes request:
    #---------------------------------------------------------------------------

    def _on_revert ( self, event ):
        """ Handles a request to revert changes.
        """
        ui = self.ui
        self._apply_context( ui._revert, ui.context )
        self._apply_context( ui._revert, ui._context )
        self.revert.Enable( False )
        ui.handler.revert( ui.info )
        ui.modified = False

    #---------------------------------------------------------------------------
    #  Handles the user interface 'modified' state changing:
    #---------------------------------------------------------------------------

    def _on_applyable ( self, state ):
        """ Handles a change to the "modified" state of the user interface .
        """
        self.apply.Enable( state )

    #---------------------------------------------------------------------------
    #  Handles editing errors:
    #---------------------------------------------------------------------------

    def _on_error ( self, errors ):
        """ Handles editing errors.
        """
        self.ok.Enable( errors == 0 )




from __future__ import absolute_import

from traits.testing.unittest_tools import unittest

from ..about_dialog import AboutDialog
from ..constant import OK, CANCEL
from ..gui import GUI
from ..toolkit import toolkit_object
from ..window import Window

ModalDialogTester = toolkit_object('util.modal_dialog_tester:ModalDialogTester')
no_modal_dialog_tester = (ModalDialogTester.__name__ == 'Unimplemented')


class TestAboutDialog(unittest.TestCase):

    def setUp(self):
        self.gui = GUI()
        self.dialog = AboutDialog()

    def test_create(self):
        # test that creation and destruction works as expected
        self.dialog._create()
        self.gui.process_events()
        self.dialog.destroy()

    def test_destroy(self):
        # test that destroy works even when no control
        self.dialog.destroy()

    def test_create_parent(self):
        # test that creation and destruction works as expected with a parent
        parent = Window()
        self.dialog.parent = parent.control
        parent._create()
        self.dialog._create()
        self.gui.process_events()
        self.dialog.destroy()
        parent.destroy()

    def test_create_ok_renamed(self):
        # test that creation and destruction works as expected with ok_label
        self.dialog.ok_label = u"Sure"
        self.dialog._create()
        self.gui.process_events()
        self.dialog.destroy()

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_accept(self):
        # test that accept works as expected
        # XXX duplicate of Dialog test, not needed?
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_run(when_opened=lambda x: x.close(accept=True))
        self.assertEqual(tester.result, OK)
        self.assertEqual(self.dialog.return_code, OK)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_close(self):
        # test that closing works as expected
        # XXX duplicate of Dialog test, not needed?
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_run(when_opened=lambda x: self.dialog.close())
        self.assertEqual(tester.result, CANCEL)
        self.assertEqual(self.dialog.return_code, CANCEL)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_ok(self):
        # test that OK works as expected
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_wait(when_opened=lambda x: x.click_button(OK))
        self.assertEqual(tester.result, OK)
        self.assertEqual(self.dialog.return_code, OK)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_renamed_ok(self):
        self.dialog.ok_label = u"Sure"
        # test that OK works as expected if renamed
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_wait(when_opened=lambda x: x.click_widget(u"Sure"))
        self.assertEqual(tester.result, OK)
        self.assertEqual(self.dialog.return_code, OK)

    @unittest.skipIf(no_modal_dialog_tester, 'ModalDialogTester unavailable')
    def test_parent(self):
        # test that lifecycle works with a parent
        parent = Window()
        self.dialog.parent = parent.control
        parent.open()
        tester = ModalDialogTester(self.dialog.open)
        tester.open_and_run(when_opened=lambda x: x.close(accept=True))
        parent.close()
        self.assertEqual(tester.result, OK)
        self.assertEqual(self.dialog.return_code, OK)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/21/2004
#
#------------------------------------------------------------------------------

""" Defines the various image enumeration editors for the wxPython user interface toolkit.
"""

#-------------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------------

import wx

from traits.api \
    import Any

# FIXME: ToolkitEditorFactory is a proxy class defined here just for backward
# compatibility. The class has been moved to the
# traitsui.editors.image_enum_editor file.
from traitsui.editors.image_enum_editor \
    import ToolkitEditorFactory

from editor \
    import Editor

from helper \
    import bitmap_cache, position_window, TraitsUIPanel

from constants \
    import WindowColor

from image_control \
    import ImageControl

#-------------------------------------------------------------------------------
#  'ReadonlyEditor' class:
#-------------------------------------------------------------------------------

class ReadonlyEditor ( Editor ):
    """ Read-only style of image enumeration editor, which displays a single
    ImageControl, representing the object trait's value.
    """
    #---------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #---------------------------------------------------------------------------

    def init ( self, parent ):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.control = ImageControl( parent, bitmap_cache( '%s%s%s' %
            ( self.factory.prefix, self.str_value, self.factory.suffix ),
            False, self.factory._image_path ) )

    #---------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #---------------------------------------------------------------------------

    def update_editor ( self ):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        self.control.Bitmap( bitmap_cache( '%s%s%s' %
            ( self.factory.prefix, self.str_value, self.factory.suffix ),
            False, self.factory._image_path ) )

#-------------------------------------------------------------------------------
#  'SimpleEditor' class:
#-------------------------------------------------------------------------------

class SimpleEditor ( ReadonlyEditor ):
    """ Simple style of image enumeration editor, which displays an
    ImageControl, representing the object trait's value. Clicking an image
    displays a dialog box for selecting an image corresponding to a different
    value.
    """
    #---------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #---------------------------------------------------------------------------

    def init ( self, parent ):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        super( SimpleEditor, self ).init( parent )
        self.control.Selected( True )
        self.control.Handler( self.popup_editor )
        self.set_tooltip()

    #---------------------------------------------------------------------------
    #  Handles the user clicking the ImageControl to display the pop-up dialog:
    #---------------------------------------------------------------------------

    def popup_editor ( self, control ):
        """ Handles the user clicking the ImageControl to display the pop-up
            dialog.
        """
        ImageEnumDialog( self )

#-------------------------------------------------------------------------------
#  'CustomEditor' class:
#-------------------------------------------------------------------------------

class CustomEditor ( Editor ):
    """ Custom style of image enumeration editor, which displays a grid of
    ImageControls. The user can click an image to select the corresponding
    value.
    """
    #---------------------------------------------------------------------------
    #  Trait definitions:
    #---------------------------------------------------------------------------

    update_handler = Any  # Callback to call when any button clicked

    #---------------------------------------------------------------------------
    #  Finishes initializing the editor by creating the underlying toolkit
    #  widget:
    #---------------------------------------------------------------------------

    def init ( self, parent ):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self._create_image_grid( parent )

    #---------------------------------------------------------------------------
    #  Populates a specified window with a grid of image buttons:
    #---------------------------------------------------------------------------

    def _create_image_grid ( self, parent ):
        """ Populates a specified window with a grid of image buttons.
        """
        # Create the panel to hold the ImageControl buttons:
        self.control = panel = TraitsUIPanel( parent, -1 )

        # Create the main sizer:
        if self.factory.cols > 1:
           sizer = wx.GridSizer( 0, self.factory.cols, 0, 0 )
        else:
           sizer = wx.BoxSizer( wx.VERTICAL )

        # Add the set of all possible choices:
        factory   = self.factory
        mapping   = factory._mapping
        cur_value = self.value
        for name in self.factory._names:
            value   = mapping[ name ]
            control = ImageControl( panel,
                          bitmap_cache( '%s%s%s' % ( factory.prefix, name,
                              factory.suffix ), False, factory._image_path ),
                          value == cur_value,
                          self.update_object )
            control.value = value
            sizer.Add( control, 0, wx.ALL, 2 )
            self.set_tooltip( control )

        # Finish setting up the control layout:
        panel.SetSizerAndFit( sizer )

    #---------------------------------------------------------------------------
    #  Handles the user clicking on an ImageControl to set an object value:
    #---------------------------------------------------------------------------

    def update_object ( self, control ):
        """ Handles the user clicking on an ImageControl to set an object value.
        """
        self.value = control.value
        if self.update_handler is not None:
            self.update_handler()

    #---------------------------------------------------------------------------
    #  Updates the editor when the object trait changes external to the editor:
    #---------------------------------------------------------------------------

    def update_editor ( self ):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        value = self.value
        for control in self.control.GetChildren():
            control.Selected( value == control.value )

#-------------------------------------------------------------------------------
#  'ImageEnumDialog' class:
#-------------------------------------------------------------------------------

class ImageEnumDialog ( wx.Frame ):
    """ Dialog box for selecting an ImageControl
    """
    #---------------------------------------------------------------------------
    #  Initializes the object:
    #---------------------------------------------------------------------------

    def __init__ ( self, editor ):
        """ Initializes the object.
        """
        wx.Frame.__init__( self, editor.control, -1, '',
                           style = wx.SIMPLE_BORDER )
        self.SetBackgroundColour( WindowColor )
        wx.EVT_ACTIVATE( self, self._on_close_dialog )
        self._closed = False

        dlg_editor = CustomEditor( self,
                                   factory        = editor.factory,
                                   ui             = editor.ui,
                                   object         = editor.object,
                                   name           = editor.name,
                                   description    = editor.description,
                                   update_handler = self._close_dialog )

        dlg_editor.init( self )

        # Wrap the dialog around the image button panel:
        sizer = wx.BoxSizer( wx.VERTICAL )
        sizer.Add( dlg_editor.control )
        sizer.Fit( self )

        # Position the dialog:
        position_window( self, parent = editor.control )
        self.Show()

    #---------------------------------------------------------------------------
    #  Closes the dialog:
    #---------------------------------------------------------------------------

    def _on_close_dialog ( self, event ):
        """ Closes the dialog.
        """
        if not event.GetActive():
            self._close_dialog()

    #---------------------------------------------------------------------------
    #  Closes the dialog:
    #---------------------------------------------------------------------------

    def _close_dialog ( self ):
        """ Closes the dialog.
        """
        if not self._closed:
            self._closed = True
            self.Destroy()

### EOF #######################################################################
